from abc import ABC, abstractmethod
from typing import Callable, Self, Tuple, ClassVar
import functools
import weakref

import lucid
from lucid.types import (
    Numeric,
    _DeviceType,
    _NumPyArray,
    _MLXArray,
    _BuiltinNumeric,
    _TensorLike,
)

from lucid._backend.metal import is_gpu_op


_GradType = _NumPyArray | _MLXArray | Tuple[_NumPyArray | _MLXArray, ...]
_GradFuncType = Callable[[], _GradType]


_ReturnGradFuncPair = Tuple[_TensorLike, _GradFuncType]
_FuncOpReturnType = _ReturnGradFuncPair | Tuple[_ReturnGradFuncPair, ...]


def func_op(
    n_in: int | None,
    n_ret: int | None,
    has_gradient: bool = True,
    device: _DeviceType = "cpu",
) -> Callable:
    def decorator(forward_func: Callable[..., _FuncOpReturnType]) -> Callable:
        @functools.wraps(forward_func)
        def wrapper(op_self: Operation, *args, **kwargs) -> Tuple[_TensorLike, ...]:
            tensors: Tuple[_TensorLike, ...] = tuple()
            requires_grad = False
            is_free = True
            dtype_hint: _BuiltinNumeric | Numeric | None = None
            inplace_target: _TensorLike | None = None

            if n_in is None:
                tensor_args = args
            else:
                if len(args) < n_in:
                    raise ValueError(
                        f"Expected at least {n_in} tensor arguments, got {len(args)}"
                    )
                tensor_args = args[:n_in]

            for arg in tensor_args:
                if isinstance(arg, _TensorLike):
                    dtype_hint = arg.dtype
                    break

            for arg in tensor_args:
                tensor = lucid._check_is_tensor(arg, device=device, dtype=dtype_hint)
                tensors += (tensor,)
                requires_grad = requires_grad or tensor.requires_grad

                if tensor.is_free:
                    tensor.to(device)
                else:
                    is_free = False
                    if tensor.device != device:
                        raise RuntimeError(
                            f"{tensor.device} tensor of shape {tensor.shape} "
                            + f"passed for {device} operation"
                            + f"('{type(op_self).__name__}')."
                        )

            if op_self._inplace:
                if n_ret != 1:
                    raise ValueError("inplace op must have a single return value.")
                if not (0 <= op_self._inplace_target < len(tensors)):
                    raise ValueError("inplace_target is out of range.")

                target = tensors[op_self._inplace_target]
                if lucid.grad_enabled() and target.requires_grad and target.is_leaf:
                    raise RuntimeError(
                        "A leaf tensor with 'requires_grad=True' "
                        "cannot be subjected to inplace operations."
                    )
                inplace_target = target

                proxy = target.new_tensor()
                proxy._op = target._op
                proxy._prev = list(target._prev)
                proxy._backward_op = target._backward_op
                proxy._backward_hooks = list(target._backward_hooks)
                proxy.grad = None
                proxy._version = target._version

                if hasattr(target, "_is_free"):
                    proxy._is_free = target._is_free
                if hasattr(target, "_is_bool_tensor"):
                    proxy._is_bool_tensor = target._is_bool_tensor

                tensors = (
                    tensors[: op_self._inplace_target]
                    + (proxy,)
                    + tensors[op_self._inplace_target + 1 :]
                )

            non_tensor_args = args[n_in:] if n_in is not None else ()
            new_args = (*tensors, *non_tensor_args)

            func_return_pairs = forward_func(op_self, *new_args, **kwargs)

            tensor_refs = tuple(weakref.ref(t) for t in tensors)

            grad_enabled = lucid.grad_enabled()
            flops_enabled = lucid.flops_enabled()
            track_graph = flops_enabled or (grad_enabled and requires_grad)

            if flops_enabled:
                op_self.flops = op_self.__flops__(*new_args, **kwargs)

            if n_ret is None:
                if not isinstance(func_return_pairs, tuple):
                    raise ValueError(
                        f"{forward_func.__name__} should return multiple '_ReturnGradFuncPair'."
                    )
                num_returns = len(func_return_pairs)
            else:
                num_returns = n_ret

            if num_returns == 1:
                func_return_pairs: _FuncOpReturnType = (func_return_pairs,)
            elif op_self._inplace:
                raise ValueError("inplace op must have a single return value.")

            if op_self._inplace:
                (ret_value, grad_func) = func_return_pairs[0]
                target = inplace_target
                if target is None:
                    raise RuntimeError("Missing inplace target tensor.")
                target.data = ret_value.data

                if ret_value.dtype is bool:
                    target._is_bool_tensor = True
                    target.dtype = bool
                else:
                    target._is_bool_tensor = False
                    target.dtype = ret_value.dtype

                target._version += 1
                func_return_pairs = ((target, grad_func),)

            results: Tuple[_TensorLike, ...] = tuple()
            for result, grad_func in func_return_pairs:
                result.requires_grad = requires_grad and has_gradient and grad_enabled
                result.to(device)
                result.free() if is_free else ...
                results += (result,)

                if not track_graph:
                    continue
                result._op = op_self

                if result.requires_grad or lucid.flops_enabled():
                    result._prev = list(tensors)
                    if not result.requires_grad:
                        continue

                    result._backward_op = BackwardOperation(
                        forward_op_ref=weakref.ref(op_self),
                        grad_func=grad_func,
                        tensor_refs=tensor_refs,
                        versions=tuple(t._version for t in tensors),
                        device=device,
                    )

            if track_graph:
                try:
                    op_self.result = results if num_returns > 1 else results[0]
                except Exception:
                    pass
            else:
                try:
                    op_self.clear()
                except Exception:
                    try:
                        op_self.result = None
                    except Exception:
                        pass

            return results if num_returns > 1 else results[0]

        return wrapper

    return decorator


def unary_func_op(has_gradient: bool = True, device: _DeviceType = "cpu") -> Callable:
    return func_op(1, 1, has_gradient, device)


def binary_func_op(has_gradient: bool = True, device: _DeviceType = "cpu") -> Callable:
    return func_op(2, 1, has_gradient, device)


def poly_func_op(has_gradient: bool = True, device: _DeviceType = "cpu") -> Callable:
    return func_op(None, 1, has_gradient, device)


class Operation(ABC):
    __fallback__: ClassVar[bool] = False

    def __init__(self) -> None:
        self.result: _TensorLike | tuple[_TensorLike, ...] | None = None
        self._inplace: bool = False
        self._inplace_target: int = 0
        self._flops: int | None = None

    def clear(self) -> None:
        self.result = None

    def inplace(self, target: int = 0) -> Self:
        self._inplace = True
        self._inplace_target = target
        return self

    @abstractmethod
    def cpu(self, *args, **kwargs) -> _FuncOpReturnType: ...

    @abstractmethod
    def gpu(self, *args, **kwargs) -> _FuncOpReturnType: ...

    def __grad__(self, *args, **kwargs) -> _GradType: ...

    def __grad_cpu__(self, *args, **kwargs) -> _GradType: ...

    def __grad_gpu__(self, *args, **kwargs) -> _GradType: ...

    @property
    def flops(self) -> int:
        if self._flops is None:
            raise ValueError(f"flops counting for {self} has not been executed.")
        return self._flops

    @flops.setter
    def flops(self, val: int) -> None:
        self._flops = val

    def __flops__(self, *args, **kwargs) -> int:
        return 0

    def __call__(self, *args, **kwargs) -> _TensorLike | tuple[_TensorLike, ...]:
        if is_gpu_op(*args):
            return self.gpu(*args, **kwargs)
        return self.cpu(*args, **kwargs)


def fallback(cls: type[Operation]) -> type[Operation]:
    cls.__fallback__ = True
    return cls


class BackwardOperation:
    def __init__(
        self,
        forward_op_ref: weakref.ref[Operation] | None,
        grad_func: _GradFuncType | None,
        tensor_refs: tuple[weakref.ref[_TensorLike]],
        versions: tuple[int, ...] = (),
        device: _DeviceType | None = "cpu",
        custom_closure: Callable[[], None] | None = None,
    ) -> None:
        self.forward_op_ref = forward_op_ref
        self.grad_func = grad_func
        self.tensor_refs = tensor_refs
        self.versions = versions
        self.device = device

        self.custom_closure = custom_closure
        self.num_inputs = len(tensor_refs)

        if self.grad_func is None and self.custom_closure is None:
            raise ValueError("Either 'grad_func' or 'custom_closure' must be provided.")

        if len(tensor_refs) != len(versions):
            raise ValueError("Numbers of 'tensor_refs' and 'versions' do not match.")

    def override_grad_func(self, new_grad_func: _GradFuncType) -> None:
        if self.custom_closure is not None:
            return
        self.grad_func = new_grad_func

    def override_tensor_refs(
        self,
        new_tensor_refs: tuple[weakref.ref[_TensorLike]],
        new_versions: tuple[int, ...] | None = None,
    ) -> None:
        self.tensor_refs = new_tensor_refs
        self.num_inputs = len(new_tensor_refs)
        if new_versions is not None:
            if len(new_versions) != len(new_tensor_refs):
                raise ValueError(
                    "Numbers of 'tensor_refs' and 'versions' do not match."
                )
            self.versions = new_versions

    def __call__(self) -> None:
        live_tensors = tuple(ref() for ref in self.tensor_refs)
        if not live_tensors or any(t is None for t in live_tensors):
            return

        if any(ver != t._version for ver, t in zip(self.versions, live_tensors)):
            raise RuntimeError(f"Tensor version mismatch detected.")

        if self.custom_closure is not None:
            self.custom_closure()
            return

        if self.device is None and self.forward_op_ref is not None:
            raise RuntimeError(
                "Only 'noop' BackwardOperation can be called without device."
            )

        grads = self.grad_func()
        if self.num_inputs == 1 or not isinstance(grads, tuple):
            grads = (grads,)

        if len(grads) != len(live_tensors):
            raise ValueError(
                f"Expected {len(live_tensors)} gradients, got {len(grads)}."
            )

        for tensor, grad in zip(live_tensors, grads):
            if not tensor.requires_grad and grad is None:
                continue
            new_grad = lucid._match_grad_shape(tensor.data, grad, device=self.device)
            lucid._set_tensor_grad(tensor, new_grad)


noop = BackwardOperation(
    forward_op_ref=None, grad_func=lambda: (), tensor_refs=(), device=None
)



from typing import Literal

import lucid
from lucid._tensor import Tensor

from lucid.nn._kernel.loss import (
    cross_entropy_kernel,
    binary_cross_entropy_kernel,
    binary_cross_entropy_with_logits_kernel,
)

_ReductionType = Literal["mean", "sum"]


def _loss_reduction(loss: Tensor, reduction: _ReductionType | None) -> Tensor:
    match reduction:
        case "mean":
            return loss.mean()
        case "sum":
            return loss.sum()
        case None:
            return loss
        case _:
            raise ValueError(
                "Invalid reduction type. Choose 'mean', 'sum', or 'none'.",
            )


def _ignore_index_loss(
    loss: Tensor,
    target_int: Tensor,
    ignore_index: int,
    reduction: _ReductionType | None,
) -> Tensor:
    mask = (target_int != ignore_index).astype(lucid.Float32)
    if reduction is None:
        return loss * mask

    loss_sum = (loss * mask).sum()
    if reduction == "sum":
        return loss_sum

    valid_count = mask.sum()
    if valid_count.item() == 0:
        return lucid.zeros_like(valid_count)

    return loss_sum / valid_count


def mse_loss(
    input_: Tensor, target: Tensor, reduction: _ReductionType | None = "mean"
) -> Tensor:
    loss = (input_ - target) ** 2
    return _loss_reduction(loss, reduction)


def binary_cross_entropy(
    input_: Tensor,
    target: Tensor,
    weight: Tensor | None = None,
    reduction: _ReductionType | None = "mean",
    eps: float = 1e-7,
) -> Tensor:
    has_weight = weight is not None
    if weight is None:
        weight = lucid.ones_like(input_, device=input_.device)

    op = binary_cross_entropy_kernel(
        reduction=reduction, eps=eps, has_weight=has_weight
    )
    return op(input_, target, weight)


def binary_cross_entropy_with_logits(
    input_: Tensor,
    target: Tensor,
    weight: Tensor | None = None,
    pos_weight: Tensor | None = None,
    reduction: _ReductionType | None = "mean",
) -> Tensor:
    has_weight = weight is not None
    has_pos_weight = pos_weight is not None
    if weight is None:
        weight = lucid.ones_like(input_, device=input_.device)
    if pos_weight is None:
        pos_weight = lucid.ones_like(input_, device=input_.device)

    op = binary_cross_entropy_with_logits_kernel(
        reduction=reduction, has_weight=has_weight, has_pos_weight=has_pos_weight
    )
    return op(input_, target, weight, pos_weight)


def cross_entropy(
    input_: Tensor,
    target: Tensor,
    weight: Tensor | None = None,
    reduction: _ReductionType | None = "mean",
    eps: float = 1e-7,
    ignore_index: int | None = None,
) -> Tensor:
    has_weight = weight is not None
    if weight is None:
        weight = lucid.ones((input_.shape[1],), device=input_.device)

    op = cross_entropy_kernel(
        reduction=reduction, eps=eps, ignore_index=ignore_index, has_weight=has_weight
    )
    return op(input_, target, weight)


def nll_loss(
    input_: Tensor,
    target: Tensor,
    weight: Tensor | None = None,
    reduction: _ReductionType | None = "mean",
    ignore_index: int | None = None,
) -> Tensor:
    target_int = target.astype(lucid.Int)
    n = input_.shape[0]
    idx = lucid.arange(n, device=input_.device).astype(lucid.Int)

    loss = -input_[idx, target_int]
    if weight is not None:
        loss *= weight[target_int]

    if ignore_index is not None:
        return _ignore_index_loss(loss, target_int, ignore_index, reduction)

    return _loss_reduction(loss, reduction)


def huber_loss(
    input_: Tensor,
    target: Tensor,
    delta: float = 1.0,
    reduction: _ReductionType | None = "mean",
) -> Tensor:
    diff = lucid.abs(input_ - target)
    quad = lucid.minimum(diff, delta)
    linear = diff - quad
    loss = 0.5 * quad**2 + delta * linear

    return _loss_reduction(loss, reduction)



from typing import Literal

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.nn.utils.rnn import PackedSequence
from lucid.types import Numeric, _DeviceType

from .activation import Tanh, ReLU


__all__ = ["RNNCell", "LSTMCell", "GRUCell", "RNNBase", "RNN", "LSTM", "GRU"]


def _get_activation(nonlinearity: str) -> type[nn.Module]:
    if nonlinearity == "tanh":
        return Tanh
    elif nonlinearity == "relu":
        return ReLU
    else:
        raise ValueError(
            f"Invalid nonlinearity '{nonlinearity}'. "
            "Supported nonlinearities are 'tanh' and 'relu'."
        )


class RNNCell(nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        bias: bool = True,
        nonlinearity: Literal["tanh", "relu"] = "tanh",
    ) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias
        self.nonlinearity = _get_activation(nonlinearity)()

        sqrt_k = 1.0 / (hidden_size**0.5)
        self.weight_ih = nn.Parameter(
            lucid.random.uniform(-sqrt_k, sqrt_k, (self.hidden_size, self.input_size))
        )
        self.weight_hh = nn.Parameter(
            lucid.random.uniform(-sqrt_k, sqrt_k, (self.hidden_size, self.hidden_size))
        )

        if self.bias:
            self.bias_ih = nn.Parameter(
                lucid.random.uniform(-sqrt_k, sqrt_k, self.hidden_size)
            )
            self.bias_hh = nn.Parameter(
                lucid.random.uniform(-sqrt_k, sqrt_k, self.hidden_size)
            )
        else:
            self.bias_ih = None
            self.bias_hh = None

    def forward(self, input_: Tensor, hx: Tensor | None = None) -> Tensor:
        if input_.ndim not in (1, 2):
            raise ValueError(
                "RNNCell expected input with 1 or 2 dimensions, "
                f"got {input_.ndim} dimensions"
            )

        is_batched = input_.ndim == 2
        if not is_batched:
            input_ = input_.unsqueeze(axis=0)
        batch_size = input_.shape[0]

        if hx is None:
            hx = lucid.zeros(
                batch_size, self.hidden_size, dtype=input_.dtype, device=input_.device
            )
        else:
            if hx.ndim not in (1, 2):
                raise ValueError(
                    "RNNCell expected hidden state with 1 or 2 dimensions, "
                    f"got {hx.ndim} dimensions"
                )
            if hx.ndim == 1:
                hx = hx.unsqueeze(axis=0)

            if hx.shape[0] != batch_size or hx.shape[1] != self.hidden_size:
                raise ValueError(
                    "RNNCell expected hidden state with shape "
                    f"({batch_size}, {self.hidden_size}), got {hx.shape}"
                )

        hy = F.linear(input_, self.weight_ih, self.bias_ih)
        hy += F.linear(hx, self.weight_hh, self.bias_hh)
        ret = self.nonlinearity(hy)

        if not is_batched:
            ret = ret.squeeze(axis=0)
        return ret


class LSTMCell(nn.Module):
    def __init__(
        self, input_size: int, hidden_size: int, bias: bool = True, **kwargs
    ) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias

        sqrt_k = 1.0 / (hidden_size**0.5)
        self.weight_ih = nn.Parameter(
            lucid.random.uniform(
                -sqrt_k, sqrt_k, (4 * self.hidden_size, self.input_size)
            )
        )
        self.weight_hh = nn.Parameter(
            lucid.random.uniform(
                -sqrt_k, sqrt_k, (4 * self.hidden_size, self.hidden_size)
            )
        )

        if self.bias:
            self.bias_ih = nn.Parameter(
                lucid.random.uniform(-sqrt_k, sqrt_k, 4 * self.hidden_size)
            )
            self.bias_hh = nn.Parameter(
                lucid.random.uniform(-sqrt_k, sqrt_k, 4 * self.hidden_size)
            )
        else:
            self.bias_ih = None
            self.bias_hh = None

    def forward(
        self, input_: Tensor, hx: tuple[Tensor, Tensor] | None = None
    ) -> tuple[Tensor, Tensor]:
        if input_.ndim not in (1, 2):
            raise ValueError(
                "LSTMCell expected input with 1 or 2 dimensions, "
                f"got {input_.ndim} dimensions"
            )

        is_batched = input_.ndim == 2
        if not is_batched:
            input_ = input_.unsqueeze(axis=0)
        batch_size = input_.shape[0]

        if hx is None:
            h_t = lucid.zeros(
                batch_size, self.hidden_size, dtype=input_.dtype, device=input_.device
            )
            c_t = lucid.zeros(
                batch_size, self.hidden_size, dtype=input_.dtype, device=input_.device
            )
        else:
            h_t, c_t = hx
            if h_t.ndim not in (1, 2) or c_t.ndim not in (1, 2):
                raise ValueError(
                    "LSTMCell expected hidden state and cell state with 1 or 2 dimensions"
                )

            if h_t.ndim == 1:
                h_t = h_t.unsqueeze(axis=0)
            if c_t.ndim == 1:
                c_t = c_t.unsqueeze(axis=0)

            if h_t.shape[0] != batch_size or h_t.shape[1] != self.hidden_size:
                raise ValueError(
                    "LSTMCell expected hidden state with shape "
                    f"({batch_size}, {self.hidden_size}), got {h_t.shape}"
                )
            if c_t.shape[0] != batch_size or c_t.shape[1] != self.hidden_size:
                raise ValueError(
                    "LSTMCell expected cell state with shape "
                    f"({batch_size}, {self.hidden_size}), got {c_t.shape}"
                )

        gates = F.linear(input_, self.weight_ih, self.bias_ih)
        gates += F.linear(h_t, self.weight_hh, self.bias_hh)

        i_t, f_t, g_t, o_t = lucid.split(gates, 4, axis=1)
        i_t = F.sigmoid(i_t)
        f_t = F.sigmoid(f_t)
        g_t = F.tanh(g_t)
        o_t = F.sigmoid(o_t)

        c_t = f_t * c_t + i_t * g_t
        h_t = o_t * F.tanh(c_t)

        if not is_batched:
            h_t = h_t.squeeze(axis=0)
            c_t = c_t.squeeze(axis=0)
        return h_t, c_t


class GRUCell(nn.Module):
    def __init__(self, input_size: int, hidden_size: int, bias: bool = True) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias

        sqrt_k = 1.0 / (hidden_size**0.5)
        self.weight_ih = nn.Parameter(
            lucid.random.uniform(
                -sqrt_k, sqrt_k, (3 * self.hidden_size, self.input_size)
            )
        )
        self.weight_hh = nn.Parameter(
            lucid.random.uniform(
                -sqrt_k, sqrt_k, (3 * self.hidden_size, self.hidden_size)
            )
        )

        if self.bias:
            self.bias_ih = nn.Parameter(
                lucid.random.uniform(-sqrt_k, sqrt_k, 3 * self.hidden_size)
            )
            self.bias_hh = nn.Parameter(
                lucid.random.uniform(-sqrt_k, sqrt_k, 3 * self.hidden_size)
            )
        else:
            self.bias_ih = None
            self.bias_hh = None

    def forward(self, input_: Tensor, hx: Tensor | None = None) -> Tensor:
        if input_.ndim not in (1, 2):
            raise ValueError(
                "GRUCell expected input with 1 or 2 dimensions, "
                f"got {input_.ndim} dimensions"
            )

        is_batched = input_.ndim == 2
        if not is_batched:
            input_ = input_.unsqueeze(axis=0)
        batch_size = input_.shape[0]

        if hx is None:
            hx = lucid.zeros(
                batch_size, self.hidden_size, dtype=input_.dtype, device=input_.device
            )
        else:
            if hx.ndim not in (1, 2):
                raise ValueError(
                    "GRUCell expected hidden state with 1 or 2 dimensions, "
                    f"got {hx.ndim} dimensions"
                )

            if hx.ndim == 1:
                hx = hx.unsqueeze(axis=0)
            if hx.shape[0] != batch_size or hx.shape[1] != self.hidden_size:
                raise ValueError(
                    "GRUCell expected hidden state with shape "
                    f"({batch_size}, {self.hidden_size}), got {hx.shape}"
                )

        input_gates = F.linear(input_, self.weight_ih, self.bias_ih)
        hidden_gates = F.linear(hx, self.weight_hh, self.bias_hh)

        i_r, i_z, i_n = lucid.split(input_gates, 3, axis=1)
        h_r, h_z, h_n = lucid.split(hidden_gates, 3, axis=1)

        r_t = F.sigmoid(i_r + h_r)
        z_t = F.sigmoid(i_z + h_z)
        n_t = F.tanh(i_n + r_t * h_n)

        h_t = (1 - z_t) * n_t + z_t * hx

        if not is_batched:
            h_t = h_t.squeeze(axis=0)
        return h_t


class RNNBase(nn.Module):
    def __init__(
        self,
        mode: Literal["RNN_TANH", "RNN_RELU", "LSTM", "GRU"],
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        bias: bool = True,
        batch_first: bool = False,
        dropout: float = 0.0,
    ) -> None:
        super().__init__()
        self.is_lstm = False
        nonlinearity = "tanh"

        if mode == "RNN_TANH":
            pass
        elif mode == "RNN_RELU":
            nonlinearity = "relu"
        elif mode == "LSTM":
            self.is_lstm = True
        elif mode == "GRU":
            pass
        else:
            raise ValueError(
                f"Invalid mode '{mode}'. Supported modes are 'RNN_TANH', "
                "'RNN_RELU', 'LSTM', or 'GRU'."
            )

        self.mode = mode
        self.nonlinearity = nonlinearity

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.bias = bias
        self.batch_first = batch_first
        self.dropout = float(dropout)

        for layer in range(num_layers):
            layer_input_size = input_size if layer == 0 else hidden_size
            sqrt_k = 1.0 / (hidden_size**0.5)

            if mode in ("RNN_TANH", "RNN_RELU"):
                w_ih = nn.Parameter(
                    lucid.random.uniform(
                        -sqrt_k, sqrt_k, (hidden_size, layer_input_size)
                    )
                )
                w_hh = nn.Parameter(
                    lucid.random.uniform(-sqrt_k, sqrt_k, (hidden_size, hidden_size))
                )
            elif mode == "LSTM":
                w_ih = nn.Parameter(
                    lucid.random.uniform(
                        -sqrt_k, sqrt_k, (4 * hidden_size, layer_input_size)
                    )
                )
                w_hh = nn.Parameter(
                    lucid.random.uniform(
                        -sqrt_k, sqrt_k, (4 * hidden_size, hidden_size)
                    )
                )
            else:
                w_ih = nn.Parameter(
                    lucid.random.uniform(
                        -sqrt_k, sqrt_k, (3 * hidden_size, layer_input_size)
                    )
                )
                w_hh = nn.Parameter(
                    lucid.random.uniform(
                        -sqrt_k, sqrt_k, (3 * hidden_size, hidden_size)
                    )
                )

            self.register_parameter(f"weight_ih_l{layer}", w_ih)
            self.register_parameter(f"weight_hh_l{layer}", w_hh)

            if bias:
                b_ih = nn.Parameter(
                    lucid.random.uniform(-sqrt_k, sqrt_k, w_ih.shape[0])
                )
                b_hh = nn.Parameter(
                    lucid.random.uniform(-sqrt_k, sqrt_k, w_hh.shape[0])
                )
                self.register_parameter(f"bias_ih_l{layer}", b_ih)
                self.register_parameter(f"bias_hh_l{layer}", b_hh)
            else:
                self.register_parameter(f"bias_ih_l{layer}", None)
                self.register_parameter(f"bias_hh_l{layer}", None)

    def _rnn_cell(
        self,
        input_: Tensor,
        hx: Tensor,
        w_ih: Tensor,
        w_hh: Tensor,
        b_ih: Tensor | None,
        b_hh: Tensor | None,
    ) -> Tensor:
        hy = F.linear(input_, w_ih, b_ih)
        hy += F.linear(hx, w_hh, b_hh)

        if self.mode == "RNN_TANH":
            return F.tanh(hy)

        return F.relu(hy)

    def _lstm_cell(
        self,
        input_: Tensor,
        hx: Tensor,
        cx: Tensor,
        w_ih: Tensor,
        w_hh: Tensor,
        b_ih: Tensor | None,
        b_hh: Tensor | None,
    ) -> tuple[Tensor, Tensor]:
        gates = F.linear(input_, w_ih, b_ih)
        gates += F.linear(hx, w_hh, b_hh)

        i_t, f_t, g_t, o_t = lucid.split(gates, 4, axis=1)
        i_t = F.sigmoid(i_t)
        f_t = F.sigmoid(f_t)
        g_t = F.tanh(g_t)
        o_t = F.sigmoid(o_t)

        c_t = f_t * cx + i_t * g_t
        h_t = o_t * F.tanh(c_t)

        return h_t, c_t

    def _gru_cell(
        self,
        input_: Tensor,
        hx: Tensor,
        w_ih: Tensor,
        w_hh: Tensor,
        b_ih: Tensor | None,
        b_hh: Tensor | None,
    ) -> Tensor:
        input_gates = F.linear(input_, w_ih, b_ih)
        hidden_gates = F.linear(hx, w_hh, b_hh)

        i_r, i_z, i_n = lucid.split(input_gates, 3, axis=1)
        h_r, h_z, h_n = lucid.split(hidden_gates, 3, axis=1)

        r_t = F.sigmoid(i_r + h_r)
        z_t = F.sigmoid(i_z + h_z)
        n_t = F.tanh(i_n + r_t * h_n)
        h_t = (1 - z_t) * n_t + z_t * hx

        return h_t

    def _init_hidden(
        self, batch_size: int, dtype: Numeric, device: _DeviceType
    ) -> Tensor | tuple[Tensor, Tensor]:
        if self.is_lstm:
            h0 = lucid.zeros(
                self.num_layers,
                batch_size,
                self.hidden_size,
                dtype=dtype,
                device=device,
            )
            c0 = lucid.zeros(
                self.num_layers,
                batch_size,
                self.hidden_size,
                dtype=dtype,
                device=device,
            )
            return h0, c0
        return lucid.zeros(
            self.num_layers, batch_size, self.hidden_size, dtype=dtype, device=device
        )

    def forward(
        self,
        input_: Tensor | PackedSequence,
        hx: Tensor | tuple[Tensor, Tensor] | None = None,
    ) -> (
        tuple[Tensor | PackedSequence, Tensor]
        | tuple[Tensor | PackedSequence, tuple[Tensor, Tensor]]
    ):
        is_packed = isinstance(input_, PackedSequence)
        if is_packed:
            data = input_.data
            batch_sizes = input_.batch_sizes
            if data.ndim != 2:
                raise ValueError(
                    "RNNBase expected packed data with 2 dimensions, "
                    f"got {data.ndim} dimensions"
                )
            if batch_sizes.ndim != 1 or batch_sizes.shape[0] == 0:
                raise ValueError(
                    "PackedSequence batch_sizes must be a non-empty 1D tensor"
                )

            batch_size = int(batch_sizes[0].item())
            feat = data.shape[1]
            if feat != self.input_size:
                raise ValueError(
                    f"RNNBase expected input with feature size {self.input_size}, got {feat}"
                )
        else:
            if input_.ndim != 3:
                raise ValueError(
                    f"RNNBase expected input with 3 dimensions, got {input_.ndim} dimensions"
                )

            if self.batch_first:
                input_ = input_.swapaxes(0, 1)

            seq_len, batch_size, feat = input_.shape
            if feat != self.input_size:
                raise ValueError(
                    f"RNNBase expected input with feature size {self.input_size}, got {feat}"
                )

        if self.is_lstm:
            if hx is None:
                hx = self._init_hidden(batch_size, input_.dtype, input_.device)
            if not (
                isinstance(hx, (tuple, list))
                and len(hx) == 2
                and isinstance(hx[0], Tensor)
                and isinstance(hx[1], Tensor)
            ):
                raise ValueError("LSTM expects hx as a tuple of (h_0, c_0)")

            h0, c0 = hx
            if h0.ndim == 2:
                h0 = h0.unsqueeze(axis=0)
            if c0.ndim == 2:
                c0 = c0.unsqueeze(axis=0)

            if h0.ndim != 3 or c0.ndim != 3:
                raise ValueError("LSTM expects h_0 and c_0 with 3 dimensions")
            if h0.shape[0] != self.num_layers or c0.shape[0] != self.num_layers:
                raise ValueError("Incorrect number of layers in h_0 or c_0")
            if h0.shape[1] != batch_size or c0.shape[1] != batch_size:
                raise ValueError("Incorrect batch size in h_0 or c_0")
            if h0.shape[2] != self.hidden_size or c0.shape[2] != self.hidden_size:
                raise ValueError("Incorrect hidden size in h_0 or c_0")

            hx_h, hx_c = h0, c0

        else:
            if hx is None:
                hx = self._init_hidden(batch_size, input_.dtype, input_.device)
            if hx.ndim == 2:
                hx = hx.unsqueeze(axis=0)
            if hx.ndim != 3:
                raise ValueError(
                    f"RNNBase expected hidden state with 3 dimensions, got {hx.ndim} dimensions"
                )

            if hx.shape[0] != self.num_layers or hx.shape[1] != batch_size:
                raise ValueError("hx has incorrect shape")
            if hx.shape[2] != self.hidden_size:
                raise ValueError("Incorrect hidden size in hx")

        layer_input = data if is_packed else input_
        h_n_list: list[Tensor] = []
        c_n_list: list[Tensor] | None = [] if self.is_lstm else None

        for layer_idx in range(self.num_layers):
            w_ih = getattr(self, f"weight_ih_l{layer_idx}")
            w_hh = getattr(self, f"weight_hh_l{layer_idx}")
            b_ih = getattr(self, f"bias_ih_l{layer_idx}")
            b_hh = getattr(self, f"bias_hh_l{layer_idx}")

            if self.is_lstm:
                h_t = hx_h[layer_idx]
                c_t = hx_c[layer_idx]
            else:
                h_t = hx[layer_idx]

            outputs = []
            if is_packed:
                final_h: list[Tensor] = []
                final_c: list[Tensor] | None = [] if self.is_lstm else None
                offset = 0

                prev_bs: int | None = None
                max_len = int(batch_sizes.shape[0])
                for t in range(max_len):
                    bs = int(batch_sizes[t].item())
                    if bs == 0:
                        break

                    if prev_bs is None:
                        prev_bs = bs
                    if bs > prev_bs:
                        raise ValueError(
                            "PackedSequence batch_sizes must be non-increasing"
                        )

                    if bs < prev_bs:
                        final_h.append(h_t[bs:prev_bs])
                        if self.is_lstm and final_c is not None:
                            final_c.append(c_t[bs:prev_bs])

                        h_t = h_t[:bs]
                        if self.is_lstm:
                            c_t = c_t[:bs]

                    step_input = layer_input[offset : offset + bs]
                    offset += bs

                    if self.is_lstm:
                        h_t, c_t = self._lstm_cell(
                            step_input, h_t, c_t, w_ih, w_hh, b_ih, b_hh
                        )
                    elif self.mode == "GRU":
                        h_t = self._gru_cell(step_input, h_t, w_ih, w_hh, b_ih, b_hh)
                    else:
                        h_t = self._rnn_cell(step_input, h_t, w_ih, w_hh, b_ih, b_hh)

                    outputs.append(h_t)
                    prev_bs = bs

                final_h.append(h_t)
                if self.is_lstm and final_c is not None:
                    final_c.append(c_t)

                h_n_list.append(
                    lucid.concatenate(tuple(reversed(final_h)), axis=0).unsqueeze(
                        axis=0
                    )
                )
                if self.is_lstm and final_c is not None and c_n_list is not None:
                    c_n_list.append(
                        lucid.concatenate(tuple(reversed(final_c)), axis=0).unsqueeze(
                            axis=0
                        )
                    )

                layer_output = (
                    lucid.concatenate(tuple(outputs), axis=0)
                    if outputs
                    else layer_input[:0]
                )

            else:
                for t in range(seq_len):
                    step_input = layer_input[t]
                    if self.is_lstm:
                        h_t, c_t = self._lstm_cell(
                            step_input, h_t, c_t, w_ih, w_hh, b_ih, b_hh
                        )
                        outputs.append(h_t.unsqueeze(axis=0))
                    elif self.mode == "GRU":
                        h_t = self._gru_cell(step_input, h_t, w_ih, w_hh, b_ih, b_hh)
                        outputs.append(h_t.unsqueeze(axis=0))
                    else:
                        h_t = self._rnn_cell(step_input, h_t, w_ih, w_hh, b_ih, b_hh)
                        outputs.append(h_t.unsqueeze(axis=0))

                layer_output = lucid.concatenate(tuple(outputs), axis=0)

            if self.training and self.dropout > 0.0 and layer_idx < self.num_layers - 1:
                layer_output = F.dropout(layer_output, p=self.dropout)

            if not is_packed:
                h_n_list.append(h_t.unsqueeze(axis=0))
                if self.is_lstm and c_n_list is not None:
                    c_n_list.append(c_t.unsqueeze(axis=0))
            layer_input = layer_output

        if is_packed:
            output = PackedSequence(
                data=layer_input,
                batch_sizes=batch_sizes,
                sorted_indices=input_.sorted_indices,
                unsorted_indices=input_.unsorted_indices,
            )
        else:
            output = layer_input

        h_n = lucid.concatenate(tuple(h_n_list), axis=0)
        if self.is_lstm and c_n_list is not None:
            c_n = lucid.concatenate(tuple(c_n_list), axis=0)

        if is_packed:
            if input_.unsorted_indices is not None:
                h_n = h_n[:, input_.unsorted_indices]
                if self.is_lstm and c_n_list is not None:
                    c_n = c_n[:, input_.unsorted_indices]
        else:
            if self.batch_first:
                output = output.swapaxes(0, 1)

        if self.is_lstm and c_n_list is not None:
            return output, (h_n, c_n)
        return output, h_n


class RNN(RNNBase):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        nonlinearity: Literal["tanh", "relu"] = "tanh",
        bias: bool = True,
        batch_first: bool = False,
        dropout: float = 0.0,
    ) -> None:
        if nonlinearity == "tanh":
            mode = "RNN_TANH"
        elif nonlinearity == "relu":
            mode = "RNN_RELU"
        else:
            raise ValueError(
                f"Invalid nonlinearity '{nonlinearity}'. "
                "Supported nonlinearities are 'tanh' and 'relu'."
            )

        super().__init__(
            mode=mode,
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bias=bias,
            batch_first=batch_first,
            dropout=dropout,
        )


class LSTM(RNNBase):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        bias: bool = True,
        batch_first: bool = False,
        dropout: float = 0.0,
    ) -> None:
        mode = "LSTM"
        super().__init__(
            mode=mode,
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bias=bias,
            batch_first=batch_first,
            dropout=dropout,
        )


class GRU(RNNBase):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        bias: bool = True,
        batch_first: bool = False,
        dropout: float = 0.0,
    ) -> None:
        mode = "GRU"
        super().__init__(
            mode=mode,
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bias=bias,
            batch_first=batch_first,
            dropout=dropout,
        )



from functools import reduce
import numpy as np

from lucid._backend.metal import mx
from lucid._tensor import Tensor
from lucid.types import _Scalar


def _calculate_fan_in_and_fan_out(tensor: Tensor) -> tuple[int, int]:
    if tensor.ndim == 2:
        fan_in = tensor.shape[1]
        fan_out = tensor.shape[0]

    elif tensor.ndim in {3, 4, 5}:
        kernel_prod = reduce(lambda x, y: x * y, tensor.shape[2:], 1)
        fan_in = tensor.shape[1] * kernel_prod
        fan_out = tensor.shape[0] * kernel_prod

    else:
        raise ValueError(
            f"Tensor with dims {tensor.ndim} is not supported. "
            + "Must be at least 2D."
        )

    return fan_in, fan_out


def _assign_like(tensor: Tensor, data: np.ndarray) -> None:
    tensor._version += 1
    if tensor.is_cpu():
        tensor.data = np.asarray(data, dtype=tensor.data.dtype)
    else:
        tensor.data = mx.array(data, dtype=tensor.data.dtype)


def uniform(tensor: Tensor, a: _Scalar, b: _Scalar) -> None:
    data = np.random.uniform(a, b, size=tensor.shape)
    _assign_like(tensor, data)


def normal(tensor: Tensor, mean: _Scalar, std: _Scalar) -> None:
    data = np.random.normal(mean, std, size=tensor.shape)
    _assign_like(tensor, data)


def constant(tensor: Tensor, val: _Scalar) -> None:
    data = np.full(tensor.shape, val)
    _assign_like(tensor, data)


def xavier_uniform(tensor: Tensor, gain: _Scalar) -> None:
    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    bound = (6 / (fan_in + fan_out)) ** 0.5 * gain
    data = np.random.uniform(-bound, bound, size=tensor.shape)
    _assign_like(tensor, data)


def xavier_normal(tensor: Tensor, gain: _Scalar) -> None:
    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    std = (2 / (fan_in + fan_out)) ** 0.5 * gain
    data = np.random.normal(0.0, std, size=tensor.shape)
    _assign_like(tensor, data)


def kaiming_uniform(tensor: Tensor, mode: str) -> None:
    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    fan = fan_in if mode == "fan_in" else fan_out
    bound = (6 / fan) ** 0.5
    data = np.random.uniform(-bound, bound, size=tensor.shape)
    _assign_like(tensor, data)


def kaiming_normal(tensor: Tensor, mode: str) -> None:
    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    fan = fan_in if mode == "fan_in" else fan_out
    std = (2 / fan) ** 0.5
    data = np.random.normal(0.0, std, size=tensor.shape)
    _assign_like(tensor, data)



from pathlib import Path
from typing import Any

from lucid.data.tokenizers import SpecialTokens, Tokenizer

from lucid._backend._C.tokenizers.bert.core import _C_BERTTokenizer


__all__ = ["BERTTokenizerFast"]


class BERTTokenizerFast(Tokenizer):
    def __init__(
        self,
        vocab: dict[str, int] | None = None,
        vocab_file: Path | str | None = None,
        unk_token: SpecialTokens | str = SpecialTokens.UNK,
        pad_token: SpecialTokens | str = SpecialTokens.PAD,
        cls_token: str = "[CLS]",
        mask_token: str = "[MASK]",
        lowercase: bool = True,
        wordpieces_prefix: str = "##",
        max_input_chars_per_word: int = 100,
        clean_text: bool = True,
    ) -> None:
        super().__init__(
            unk_token=unk_token,
            pad_token=pad_token,
            bos_token=cls_token,
            eos_token="[SEP]",
        )

        self.cls_token = cls_token
        self.sep_token = "[SEP]"
        self.mask_token = mask_token
        self.lowercase = lowercase
        self.wordpieces_prefix = wordpieces_prefix
        self.max_input_chars_per_word = max_input_chars_per_word
        self.clean_text = clean_text

        self._backend = _C_BERTTokenizer(
            vocab=vocab,
            vocab_file=Path(vocab_file) if vocab_file is not None else None,
            unk_token=self.unk_token,
            pad_token=self.pad_token,
            cls_token=self.cls_token,
            mask_token=self.mask_token,
            wordpieces_prefix=self.wordpieces_prefix,
            max_input_chars_per_word=self.max_input_chars_per_word,
            lowercase=self.lowercase,
            clean_text=self.clean_text,
        )

    @property
    def vocab_size(self) -> int:
        size = getattr(self._backend, "vocab_size")
        return int(size() if callable(size) else size)

    def tokenize(self, text: str) -> list[str]:
        return list(self._backend.tokenize(text))

    def convert_tokens_to_ids(self, tokens: str | list[str]) -> int | list[int]:
        if isinstance(tokens, str):
            return int(self._backend.convert_token_to_id(tokens))
        return [int(x) for x in self._backend.convert_tokens_to_ids(tokens)]

    def convert_ids_to_tokens(self, ids: int | list[int]) -> str | list[str]:
        if isinstance(ids, int):
            return str(self._backend.convert_id_to_token(ids))
        return [str(x) for x in self._backend.convert_ids_to_tokens(ids)]

    def convert_tokens_to_string(self, tokens: list[str]) -> str:
        return str(self._backend.convert_tokens_to_string(tokens))

    def build_inputs_with_special_tokens(self, tokens: list[str]) -> list[str]:
        return list(self._backend.build_inputs_with_special_tokens(tokens))

    def encode_plus(
        self,
        text_a: str,
        text_b: str | None = None,
    ) -> dict[str, list[int]]:
        out = self._backend.encode_plus(text_a, text_b)
        return {
            "input_ids": [int(x) for x in out.input_ids],
            "token_type_ids": [int(x) for x in out.token_type_ids],
            "attention_mask": [int(x) for x in out.attention_mask],
        }

    def save_pretrained(self, save_directory: Path | str) -> list[str]:
        save_path = Path(save_directory)
        save_path.mkdir(parents=True, exist_ok=True)

        vocab_path = save_path / "vocab.txt"
        with vocab_path.open("w", encoding="utf-8") as f:
            for idx in range(self.vocab_size):
                f.write(self.convert_ids_to_tokens(idx) + "\n")

        config = {
            "unk_token": self.unk_token,
            "pad_token": self.pad_token,
            "cls_token": self.cls_token,
            "mask_token": self.mask_token,
            "lowercase": self.lowercase,
            "wordpieces_prefix": self.wordpieces_prefix,
            "max_input_chars_per_word": self.max_input_chars_per_word,
            "clean_text": self.clean_text,
        }
        config_path = self._save_tokenizer_config(save_path, config)
        return [str(vocab_path), str(config_path)]

    @classmethod
    def from_pretrained(
        cls, pretrained_model_name_or_path: str | Path, **kwrags: Any
    ) -> BERTTokenizerFast:
        path = Path(pretrained_model_name_or_path)
        vocab_file = path / "vocab.txt" if path.is_dir() else path
        if not vocab_file.exists():
            raise FileNotFoundError(f"Cannot find vocabulary file: {vocab_file}")

        allowed = {
            "unk_token",
            "pad_token",
            "cls_token",
            "mask_token",
            "lowercase",
            "wordpieces_prefix",
            "max_input_chars_per_word",
            "clean_text",
        }
        init_kwargs = cls._load_tokenizer_config(path, allowed_keys=allowed)
        init_kwargs.update(kwrags)
        return cls(vocab_file=vocab_file, **init_kwargs)



import math

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.types import _Scalar


__all__ = ["ScaledDotProductAttention", "MultiHeadAttention"]


@nn.auto_repr("dropout_p", "is_causal", "scale")
class ScaledDotProductAttention(nn.Module):
    def __init__(
        self,
        attn_mask: Tensor | None = None,
        dropout_p: float = 0.0,
        is_causal: bool = False,
        scale: _Scalar | None = None,
    ) -> None:
        super().__init__()
        self.attn_mask = attn_mask
        self.dropout_p = dropout_p
        self.is_causal = is_causal
        self.scale = scale

    def forward(self, query: Tensor, key: Tensor, value: Tensor) -> Tensor:
        return F.scaled_dot_product_attention(
            query,
            key,
            value,
            self.attn_mask,
            self.dropout_p,
            self.is_causal,
            self.scale,
        )


@nn.auto_repr(
    "embed_dim",
    "num_heads",
    "dropout",
    "bias",
    "use_separate_proj_weight",
    "add_bias_kv",
    "add_zero_attn",
)
class MultiHeadAttention(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        dropout: float = 0.0,
        bias: bool = True,
        use_separate_proj_weight: bool = True,
        add_bias_kv: bool = False,
        add_zero_attn: bool = False,
        kdim: int | None = None,
        vdim: int | None = None,
    ) -> None:
        super().__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.dropout = dropout
        self.use_separate_proj_weight = use_separate_proj_weight
        self.add_bias_kv = add_bias_kv
        self.add_zero_attn = add_zero_attn

        self.head_dim = embed_dim // num_heads
        if self.head_dim * num_heads != embed_dim:
            raise ValueError("embed_dim must be divisible by num_heads.")

        kdim = kdim if kdim is not None else embed_dim
        vdim = vdim if vdim is not None else embed_dim

        if use_separate_proj_weight:
            self.q_proj = nn.Linear(embed_dim, embed_dim, bias=bias)
            self.k_proj = nn.Linear(kdim, embed_dim, bias=bias)
            self.v_proj = nn.Linear(vdim, embed_dim, bias=bias)

            self.register_parameter("in_proj_weight", None)
            self.register_parameter("in_proj_bias", None)
        else:
            if kdim != embed_dim or vdim != embed_dim:
                raise ValueError(
                    "in_proj_weight requires kdim and vdim to equal embed_dim."
                )

            weight_ = lucid.empty(3 * embed_dim, embed_dim)
            self.in_proj_weight = nn.Parameter(weight_)
            if bias:
                bias_ = lucid.empty(3 * embed_dim)
                self.in_proj_bias = nn.Parameter(bias_)
            else:
                self.register_parameter("in_proj_bias", None)

            self.q_proj = None
            self.k_proj = None
            self.v_proj = None

        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias)

        if add_bias_kv:
            self.bias_k = nn.Parameter(lucid.zeros(1, 1, embed_dim))
            self.bias_v = nn.Parameter(lucid.zeros(1, 1, embed_dim))
        else:
            self.bias_k = None
            self.bias_v = None

        self.scale: _Scalar = self.head_dim**-0.5
        self._reset_parameters()

    def _reset_parameters(self) -> None:
        if self.in_proj_weight is None:
            return

        nn.init.kaiming_uniform(self.in_proj_weight)
        if self.in_proj_bias is not None:
            fan_in, _ = nn.init._dist._calculate_fan_in_and_fan_out(self.in_proj_weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform(self.in_proj_bias, -bound, bound)

    def forward(
        self,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        key_padding_mask: Tensor | None = None,
        attn_mask: Tensor | None = None,
        is_causal: bool = False,
        kv_cache: nn.KVCache | None = None,
        use_cache: bool = False,
        cache_position: Tensor | None = None,
        cache_layer_idx: int | None = None,
    ) -> Tensor:
        N, q_len = query.shape[:2]
        k_len, v_len = key.shape[1], value.shape[1]

        if self.in_proj_weight is None:
            q = self.q_proj(query)
            k = self.k_proj(key)
            v = self.v_proj(value)
        else:
            if query is key and key is value:
                qkv = F.linear(query, self.in_proj_weight, self.in_proj_bias)
                q, k, v = lucid.chunk(qkv, 3, axis=-1)
            else:
                w_q, w_k, w_v = lucid.chunk(self.in_proj_weight, 3, axis=0)
                if self.in_proj_bias is not None:
                    b_q, b_k, b_v = lucid.chunk(self.in_proj_bias, 3, axis=0)
                else:
                    b_q = b_k = b_v = None

                q = F.linear(query, w_q, b_q)
                k = F.linear(key, w_k, b_k)
                v = F.linear(value, w_v, b_v)

        q = q.reshape(N, self.num_heads, q_len, self.head_dim)
        k = k.reshape(N, self.num_heads, k_len, self.head_dim)
        v = v.reshape(N, self.num_heads, v_len, self.head_dim)

        if use_cache:
            if kv_cache is None:
                raise ValueError("use_cache=True requires kv_cache.")
            if cache_layer_idx is None:
                raise ValueError("use_cache=True requires cache_layer_idx.")
            k, v = kv_cache.update(k, v, cache_layer_idx, cache_position)

        if self.add_bias_kv:
            bias_k = self.bias_k.reshape(1, self.num_heads, 1, self.head_dim).repeat(
                N, axis=0
            )
            bias_v = self.bias_v.reshape(1, self.num_heads, 1, self.head_dim).repeat(
                N, axis=0
            )
            k = lucid.concatenate([k, bias_k], axis=2)
            v = lucid.concatenate([v, bias_v], axis=2)

            if attn_mask is not None:
                attn_mask = lucid.pad(attn_mask, (0, 1))

        if self.add_zero_attn:
            zeros = lucid.zeros(N, self.num_heads, 1, self.head_dim, dtype=q.dtype)
            k = lucid.concatenate([k, zeros], axis=2)
            v = lucid.concatenate([v, zeros], axis=2)

            if attn_mask is not None:
                attn_mask = lucid.pad(attn_mask, (0, 1))

        if key_padding_mask is not None:
            key_padding_mask = key_padding_mask[:, None, None, :]
            attn_mask = (
                key_padding_mask * -1e12
                if attn_mask is None
                else attn_mask + key_padding_mask * -1e12
            )

        attn_output = F.scaled_dot_product_attention(
            q, k, v, attn_mask, self.dropout, is_causal=is_causal, scale=self.scale
        )
        attn_output = attn_output.reshape(N, q_len, self.embed_dim)

        output = self.out_proj(attn_output)
        return output



from lucid._tensor.tensor import Tensor
from lucid._backend.core import _GradType

from lucid._func import ufunc
from lucid._util import func as util_func

from .base import FusedBackwardOp


__all__ = [
    "DoubleNeg",
    "DoubleReciprocal",
    "LogExp",
    "DoubleT",
    "DoubleMT",
    "DoubleReshape",
    "SqueezeUnsqueeze",
    "UnsqueezeSqueeze",
]


class _IdentityFusion(FusedBackwardOp):
    @classmethod
    def __grad__(cls, rets: tuple[Tensor]) -> _GradType:
        return rets[0].grad


class _IdentityViewFusion(FusedBackwardOp):
    @classmethod
    def __grad__(cls, ins: tuple[Tensor], rets: tuple[Tensor]) -> _GradType:
        v = rets[0]
        x = ins[0]
        return v.grad.reshape(x.shape) if v.grad is not None else None


class DoubleNeg(_IdentityFusion):
    op1 = ufunc._neg
    op2 = ufunc._neg


class DoubleReciprocal(_IdentityFusion):
    op1 = ufunc.reciprocal
    op2 = ufunc.reciprocal


class LogExp(_IdentityFusion):
    op1 = ufunc.exp
    op2 = ufunc.log

    heuristic_thresh = 10_000


class DoubleT(_IdentityFusion):
    op1 = ufunc._T
    op2 = ufunc._T


class DoubleMT(_IdentityFusion):
    op1 = ufunc._mT
    op2 = ufunc._mT


class DoubleReshape(_IdentityViewFusion):
    op1 = util_func.reshape
    op2 = util_func.reshape


class _DoubleReshapeImmediate(_IdentityViewFusion):
    op1 = getattr(util_func, "_reshape_immediate", None)
    op2 = getattr(util_func, "_reshape_immediate", None)


class SqueezeUnsqueeze(_IdentityViewFusion):
    op1 = util_func.squeeze
    op2 = util_func.unsqueeze


class UnsqueezeSqueeze(_IdentityViewFusion):
    op1 = util_func.unsqueeze
    op2 = util_func.squeeze



from functools import partial
from types import ModuleType

import numpy as np

from lucid._backend.core import Operation, func_op, _FuncOpReturnType, _GradType
from lucid._backend.metal import mx
from lucid._tensor import Tensor


class embedding_kernel(Operation):
    def __init__(self, padding_idx: int = -1) -> None:
        super().__init__()
        self.padding_idx = int(padding_idx)
        self._indices = None
        self._num_embeddings = None

    def clear(self) -> None:
        super().clear()
        self.padding_idx = -1
        self._indices = None
        self._num_embeddings = None

    @func_op(n_in=2, n_ret=1)
    def cpu(self, indices: Tensor, weight: Tensor) -> _FuncOpReturnType:
        return self._forward(indices, weight, lib_=np)

    @func_op(n_in=2, n_ret=1, device="gpu")
    def gpu(self, indices: Tensor, weight: Tensor) -> _FuncOpReturnType:
        return self._forward(indices, weight, lib_=mx)

    def _forward(
        self, indices: Tensor, weight: Tensor, lib_: ModuleType
    ) -> _FuncOpReturnType:
        idx = indices.data
        out = weight.data[idx]

        self._indices = idx
        self._num_embeddings = int(weight.shape[0])

        self.result = Tensor(out)
        return self.result, partial(self.__grad__, lib_=lib_)

    def __grad__(self, lib_: ModuleType) -> _GradType:
        if self.result is None or self.result.grad is None:
            raise RuntimeError("embedding backward called before forward.")

        if self._indices is None or self._num_embeddings is None:
            raise RuntimeError("embedding cached data missing.")

        grad_out = self.result.grad
        idx = self._indices.reshape(-1)
        grad_flat = grad_out.reshape(idx.shape[0], -1)

        if lib_ is np:
            if self.padding_idx >= 0:
                keep = idx != self.padding_idx
                idx = idx[keep]
                grad_flat = grad_flat[keep]

            grad_w = np.zeros(
                (self._num_embeddings, grad_flat.shape[1]), dtype=grad_out.dtype
            )
            np.add.at(grad_w, idx, grad_flat)

        else:
            grad_w = mx.zeros(
                (self._num_embeddings, grad_flat.shape[1]), dtype=grad_out.dtype
            )
            for i in range(idx.shape[0]):
                if self.padding_idx >= 0 and int(idx[i]) == self.padding_idx:
                    continue
                grad_w = grad_w.at[idx[i]].add(grad_flat[i])

        return None, grad_w



from abc import ABC, abstractmethod
from typing import Any, override

import lucid
from lucid._tensor import Tensor


__all__ = ["Cache", "KVCache", "DynamicKVCache", "StaticKVCache", "EncoderDecoderCache"]


class Cache(ABC):
    @abstractmethod
    def reset(self) -> None: ...

    def reorder_cache(self, beam_idx: Tensor) -> None:
        self.batch_select_indices(beam_idx)

    @abstractmethod
    def batch_select_indices(self, indices: Tensor) -> None: ...

    @abstractmethod
    def batch_repeat_interleave(self, repeats: int) -> None: ...

    @abstractmethod
    def crop(self, max_length: int) -> None: ...

    def update(self, *args, **kwargs) -> Any: ...

    def get_max_cache_shape(self) -> int | None:
        return None


class KVCache(Cache):
    def __init__(self) -> None:
        super().__init__()
        self.key_cache: list[Tensor | None]
        self.value_cache: list[Tensor | None]

    def _check_valid_layer_idx(self, layer_idx: int, dynamic: bool = False) -> None:
        if layer_idx < 0:
            raise ValueError(f"Invalid layer_idx '{layer_idx}'")
        if dynamic:
            return
        if layer_idx >= len(self.key_cache):
            raise ValueError(
                f"layer_idx '{layer_idx}' bigger "
                f"than cache length '{len(self.key_cache)}'"
            )

    def _check_key_value_shape(self, key: Tensor, value: Tensor) -> None:
        if key.shape != value.shape:
            raise ValueError(
                f"key and value has different shapes: {key.shape} != {value.shape}"
            )

    @override
    @abstractmethod
    def update(
        self,
        key: Tensor,
        value: Tensor,
        layer_idx: int,
        cache_position: Tensor | None = None,
    ) -> tuple[Tensor, Tensor]: ...

    @abstractmethod
    def get(self, layer_idx: int) -> tuple[Tensor, Tensor] | None: ...

    @abstractmethod
    def get_seq_length(self, layer_idx: int = 0) -> int: ...

    @abstractmethod
    def _crop_layer(self, layer_idx: int, max_length: int) -> None: ...

    def _coerce_index_device(self, index: Tensor, device: str) -> Tensor:
        if index.device == device:
            return index
        if index.is_free:
            return index.to(device)
        raise lucid.DeviceMismatchError(to=device, from_=index.device)

    def batch_select_indices(self, indices: Tensor) -> None:
        if indices.ndim != 1:
            raise ValueError(
                "indices must be 1-D for cache selection "
                f"(got indices.ndim={indices.ndim})."
            )

        for layer_idx in range(len(self.key_cache)):
            k = self.key_cache[layer_idx]
            v = self.value_cache[layer_idx]
            if k is None or v is None:
                continue

            cur_indices = self._coerce_index_device(indices, k.device)
            self.key_cache[layer_idx] = k[cur_indices, ...]
            self.value_cache[layer_idx] = v[cur_indices, ...]

    def batch_repeat_interleave(self, repeats: int) -> None:
        if repeats < 1:
            raise ValueError(f"repeats must be >= 1 (got {repeats}).")
        if repeats == 1:
            return

        for layer_idx in range(len(self.key_cache)):
            k = self.key_cache[layer_idx]
            v = self.value_cache[layer_idx]
            if k is None or v is None:
                continue

            self.key_cache[layer_idx] = k.repeat(repeats, axis=0)
            self.value_cache[layer_idx] = v.repeat(repeats, axis=0)

    def crop(self, max_length: int) -> None:
        if max_length < 0:
            raise ValueError(f"max_length must be >= 0 (got {max_length}).")

        for layer_idx in range(len(self.key_cache)):
            self._crop_layer(layer_idx, max_length)


class DynamicKVCache(KVCache):
    def __init__(self) -> None:
        super().__init__()
        self.key_cache: list[Tensor | None] = []
        self.value_cache: list[Tensor | None] = []
        self._seq_lens: list[int] = []

    def update(
        self,
        key: Tensor,
        value: Tensor,
        layer_idx: int,
        cache_position: Tensor | None = None,
    ) -> tuple[Tensor, Tensor]:
        self._check_valid_layer_idx(layer_idx, dynamic=True)
        self._check_key_value_shape(key, value)

        if key.ndim < 2:
            raise ValueError(
                "KV cache expects key/value to have at least 2 dimensions "
                f"(got key.ndim={key.ndim})."
            )

        while len(self.key_cache) <= layer_idx:
            self.key_cache.append(None)
            self.value_cache.append(None)
            self._seq_lens.append(0)

        past_k = self.key_cache[layer_idx]
        past_v = self.value_cache[layer_idx]

        if cache_position is None:
            if past_k is None or past_v is None:
                new_k, new_v = key, value
            else:
                new_k = lucid.concatenate([past_k, key], axis=-2)
                new_v = lucid.concatenate([past_v, value], axis=-2)
        else:
            if cache_position.device != key.device:
                if cache_position.is_free:
                    cache_position = cache_position.to(key.device)
                else:
                    raise lucid.DeviceMismatchError(
                        to=key.device, from_=cache_position.device
                    )

            min_pos = int(lucid.min(cache_position).item())
            max_pos = int(lucid.max(cache_position).item())
            if min_pos < 0:
                raise ValueError(f"cache_position must be >= 0 (got {min_pos}).")

            target_len = max_pos + 1
            cur_len = self._seq_lens[layer_idx]

            if past_k is None or past_v is None:
                cache_shape = list(key.shape)
                cache_shape[-2] = target_len
                cache_shape_t = tuple(cache_shape)

                new_k = lucid.zeros(cache_shape_t, dtype=key.dtype, device=key.device)
                new_v = lucid.zeros(
                    cache_shape_t, dtype=value.dtype, device=value.device
                )
            else:
                expected_shape = past_k.shape[:-2] + past_k.shape[-1:]
                incoming_shape = key.shape[:-2] + key.shape[-1:]

                if expected_shape != incoming_shape:
                    raise ValueError(
                        "DynamicKVCache shape mismatch on non-seq dimensions: "
                        f"expected {expected_shape}, got {incoming_shape}"
                    )

                if cur_len >= target_len:
                    new_k, new_v = past_k, past_v
                else:
                    grow_shape = list(past_k.shape)
                    grow_shape[-2] = target_len - cur_len
                    grow_shape_t = tuple(grow_shape)

                    pad_k = lucid.zeros(
                        grow_shape_t, dtype=past_k.dtype, device=past_k.device
                    )
                    pad_v = lucid.zeros(
                        grow_shape_t, dtype=past_v.dtype, device=past_v.device
                    )
                    new_k = lucid.concatenate([past_k, pad_k], axis=-2)
                    new_v = lucid.concatenate([past_v, pad_v], axis=-2)

            if cache_position.ndim == 0:
                if key.shape[-2] != 1:
                    raise ValueError(
                        "0-d cache_position only supports single-token updates "
                        f"(got new_tokens={key.shape[-2]})."
                    )

                pos = int(cache_position.item())
                new_k[..., pos : pos + 1, :] = key
                new_v[..., pos : pos + 1, :] = value
                self._seq_lens[layer_idx] = max(cur_len, pos + 1)

            elif cache_position.ndim == 1:
                if cache_position.shape[0] != key.shape[-2]:
                    raise ValueError(
                        "cache_position length must match key/value seq_len "
                        f"(got {cache_position.shape[0]} vs {key.shape[-2]})."
                    )

                new_k[..., cache_position, :] = key
                new_v[..., cache_position, :] = value
                self._seq_lens[layer_idx] = max(cur_len, max_pos + 1)

            elif cache_position.ndim == 2:
                if key.ndim < 3:
                    raise ValueError(
                        "2-D cache_position requires key/value to have a batch axis "
                        f"(got key.ndim={key.ndim})."
                    )
                if cache_position.shape[0] != key.shape[0]:
                    raise ValueError(
                        "cache_position batch size must match key/value batch size "
                        f"(got {cache_position.shape[0]} vs {key.shape[0]})."
                    )
                if cache_position.shape[1] != key.shape[-2]:
                    raise ValueError(
                        "cache_position second dim must match key/value seq_len "
                        f"(got {cache_position.shape[1]} vs {key.shape[-2]})."
                    )

                for batch_idx in range(key.shape[0]):
                    pos = cache_position[batch_idx]
                    for token_idx in range(key.shape[-2]):
                        token_pos = int(pos[token_idx].item())
                        new_k[batch_idx, ..., token_pos : token_pos + 1, :] = key[
                            batch_idx, ..., token_idx : token_idx + 1, :
                        ]
                        new_v[batch_idx, ..., token_pos : token_pos + 1, :] = value[
                            batch_idx, ..., token_idx : token_idx + 1, :
                        ]
                self._seq_lens[layer_idx] = max(cur_len, max_pos + 1)

            else:
                raise ValueError(
                    "Only scalar, 1-D, or 2-D cache_position is supported "
                    f"(got cache_position.ndim={cache_position.ndim})."
                )
        if cache_position is None:
            self._seq_lens[layer_idx] = new_k.shape[-2]

        self.key_cache[layer_idx] = new_k
        self.value_cache[layer_idx] = new_v

        return new_k, new_v

    def get(self, layer_idx: int) -> tuple[Tensor, Tensor] | None:
        if layer_idx >= len(self.key_cache):
            return None

        k = self.key_cache[layer_idx]
        v = self.value_cache[layer_idx]

        if k is None or v is None:
            return None
        seq_len = self._seq_lens[layer_idx]
        return k[..., :seq_len, :], v[..., :seq_len, :]

    def get_seq_length(self, layer_idx: int = 0) -> int:
        self._check_valid_layer_idx(layer_idx, dynamic=True)
        return self._seq_lens[layer_idx]

    def reset(self) -> None:
        self.key_cache.clear()
        self.value_cache.clear()
        self._seq_lens.clear()

    def _crop_layer(self, layer_idx: int, max_length: int) -> None:
        k = self.key_cache[layer_idx]
        v = self.value_cache[layer_idx]
        if k is None or v is None:
            return

        seq_len = k.shape[-2]
        if seq_len <= max_length:
            return

        if max_length == 0:
            self.key_cache[layer_idx] = k[..., :0, :]
            self.value_cache[layer_idx] = v[..., :0, :]
            self._seq_lens[layer_idx] = 0
            return

        self.key_cache[layer_idx] = k[..., -max_length:, :]
        self.value_cache[layer_idx] = v[..., -max_length:, :]
        self._seq_lens[layer_idx] = max_length


class StaticKVCache(KVCache):
    def __init__(self, max_cache_len: int, num_layers: int) -> None:
        super().__init__()
        self.max_cache_len = max_cache_len
        self.num_layers = num_layers

        self.key_cache: list[Tensor | None] = [None] * num_layers
        self.value_cache: list[Tensor | None] = [None] * num_layers
        self._seq_lens: list[int] = [0] * num_layers

    def update(
        self,
        key: Tensor,
        value: Tensor,
        layer_idx: int,
        cache_position: Tensor | None = None,
    ) -> tuple[Tensor, Tensor]:
        self._check_valid_layer_idx(layer_idx)
        self._check_key_value_shape(key, value)

        if key.ndim < 2:
            raise ValueError(
                "KV cache expects key/value to have at least 2 dimensions "
                f"(got key.ndim={key.ndim})."
            )

        new_tokens = key.shape[-2]
        if new_tokens < 1:
            raise ValueError(f"key/value seq_len must be >= 1 (got {new_tokens}).")

        cache_shape = list(key.shape)
        cache_shape[-2] = self.max_cache_len
        cache_shape_t = tuple(cache_shape)

        k_cache = self.key_cache[layer_idx]
        v_cache = self.value_cache[layer_idx]

        if k_cache is None or v_cache is None:
            k_cache = lucid.zeros(cache_shape_t, dtype=key.dtype, device=key.device)
            v_cache = lucid.zeros(cache_shape_t, dtype=value.dtype, device=value.device)

            self.key_cache[layer_idx] = k_cache
            self.value_cache[layer_idx] = v_cache

        else:
            if k_cache.shape != cache_shape_t or v_cache.shape != cache_shape_t:
                raise ValueError(
                    "StaticKVCache key/value cache shape mismatch: "
                    f"expected {cache_shape_t}, got {k_cache.shape} / {v_cache.shape}"
                )

        cur_len = self._seq_lens[layer_idx]

        if cache_position is None:
            start = cur_len
            end = start + new_tokens
            if end > self.max_cache_len:
                raise ValueError(
                    f"StaticKVCache exceeded max_cache_len={self.max_cache_len}: "
                    f"need end={end} (cur_len={cur_len}, new_tokens={new_tokens})."
                )

            k_cache[..., start:end, :] = key
            v_cache[..., start:end, :] = value
            self._seq_lens[layer_idx] = end

        else:
            if cache_position.device != key.device:
                if cache_position.is_free:
                    cache_position = cache_position.to(key.device)
                else:
                    raise lucid.DeviceMismatchError(
                        to=key.device, from_=cache_position.device
                    )

            min_pos = int(lucid.min(cache_position).item())
            max_pos = int(lucid.max(cache_position).item())

            if min_pos < 0:
                raise ValueError(f"cache_position must be >= 0 (got {min_pos}).")
            if max_pos >= self.max_cache_len:
                raise ValueError(
                    f"cache_position out of bounds for max_cache_len={self.max_cache_len}: "
                    f"max_pos={max_pos}."
                )

            if cache_position.ndim == 0:
                if new_tokens != 1:
                    raise ValueError(
                        "0-d cache_position only supports single-token updates "
                        f"(got new_tokens={new_tokens})."
                    )

                pos = int(cache_position.item())
                k_cache[..., pos : pos + 1, :] = key
                v_cache[..., pos : pos + 1, :] = value
                self._seq_lens[layer_idx] = max(cur_len, pos + 1)

            elif cache_position.ndim == 1:
                if cache_position.shape[0] != new_tokens:
                    raise ValueError(
                        "cache_position length must match key/value seq_len "
                        f"(got {cache_position.shape[0]} vs {new_tokens})."
                    )

                k_cache[..., cache_position, :] = key
                v_cache[..., cache_position, :] = value
                self._seq_lens[layer_idx] = max(cur_len, max_pos + 1)

            elif cache_position.ndim == 2:
                if key.ndim < 3:
                    raise ValueError(
                        "2-D cache_position requires key/value to have a batch axis "
                        f"(got key.ndim={key.ndim})."
                    )
                if cache_position.shape[0] != key.shape[0]:
                    raise ValueError(
                        "cache_position batch size must match key/value batch size "
                        f"(got {cache_position.shape[0]} vs {key.shape[0]})."
                    )
                if cache_position.shape[1] != new_tokens:
                    raise ValueError(
                        "cache_position second dim must match key/value seq_len "
                        f"(got {cache_position.shape[1]} vs {new_tokens})."
                    )

                for batch_idx in range(key.shape[0]):
                    pos = cache_position[batch_idx]
                    for token_idx in range(new_tokens):
                        token_pos = int(pos[token_idx].item())
                        k_cache[batch_idx, ..., token_pos : token_pos + 1, :] = key[
                            batch_idx, ..., token_idx : token_idx + 1, :
                        ]
                        v_cache[batch_idx, ..., token_pos : token_pos + 1, :] = value[
                            batch_idx, ..., token_idx : token_idx + 1, :
                        ]
                self._seq_lens[layer_idx] = max(cur_len, max_pos + 1)

            else:
                raise ValueError(
                    "Only scalar, 1-D, or 2-D cache_position is supported "
                    f"(got cache_position.ndim={cache_position.ndim})."
                )

        out = self.get(layer_idx)
        assert out is not None
        return out

    def get(self, layer_idx: int) -> tuple[Tensor, Tensor] | None:
        self._check_valid_layer_idx(layer_idx)
        k = self.key_cache[layer_idx]
        v = self.value_cache[layer_idx]

        if k is None or v is None:
            return None

        seq_len = self._seq_lens[layer_idx]
        return k[..., :seq_len, :], v[..., :seq_len, :]

    def get_seq_length(self, layer_idx: int = 0) -> int:
        self._check_valid_layer_idx(layer_idx)
        return self._seq_lens[layer_idx]

    @override
    def get_max_cache_shape(self) -> int | None:
        return self.max_cache_len

    def reset(self) -> None:
        self.key_cache = [None] * self.num_layers
        self.value_cache = [None] * self.num_layers
        self._seq_lens = [0] * self.num_layers

    def _crop_layer(self, layer_idx: int, max_length: int) -> None:
        k = self.key_cache[layer_idx]
        v = self.value_cache[layer_idx]
        if k is None or v is None:
            return

        seq_len = self._seq_lens[layer_idx]
        if seq_len <= max_length:
            return

        if max_length == 0:
            self._seq_lens[layer_idx] = 0
            return

        start = seq_len - max_length
        tail_k = k[..., start:seq_len, :].detach()
        tail_v = v[..., start:seq_len, :].detach()

        k[..., :max_length, :] = tail_k
        v[..., :max_length, :] = tail_v

        self._seq_lens[layer_idx] = max_length


class EncoderDecoderCache(Cache):
    def __init__(
        self,
        self_attention_cache: KVCache | None = None,
        cross_attention_cache: KVCache | None = None,
    ) -> None:
        self.self_attention_cache = (
            self_attention_cache
            if self_attention_cache is not None
            else DynamicKVCache()
        )
        self.cross_attention_cache = (
            cross_attention_cache
            if cross_attention_cache is not None
            else DynamicKVCache()
        )
        self.is_updated: dict[int, bool] = {}

    @staticmethod
    def _select_cache(
        self_attention_cache: KVCache,
        cross_attention_cache: KVCache,
        is_cross_attention: bool,
    ) -> KVCache:
        return cross_attention_cache if is_cross_attention else self_attention_cache

    def update(
        self,
        key: Tensor,
        value: Tensor,
        layer_idx: int,
        cache_position: Tensor | None = None,
        is_cross_attention: bool = False,
    ) -> tuple[Tensor, Tensor]:
        target_cache = self._select_cache(
            self.self_attention_cache, self.cross_attention_cache, is_cross_attention
        )
        updated = target_cache.update(key, value, layer_idx, cache_position)
        if is_cross_attention:
            self.is_updated[layer_idx] = True
        return updated

    def get(
        self,
        layer_idx: int,
        is_cross_attention: bool = False,
    ) -> tuple[Tensor, Tensor] | None:
        target_cache = self._select_cache(
            self.self_attention_cache, self.cross_attention_cache, is_cross_attention
        )
        return target_cache.get(layer_idx)

    def get_seq_length(
        self, layer_idx: int = 0, is_cross_attention: bool = False
    ) -> int:
        target_cache = self._select_cache(
            self.self_attention_cache, self.cross_attention_cache, is_cross_attention
        )
        return target_cache.get_seq_length(layer_idx)

    def reset(self) -> None:
        self.self_attention_cache.reset()
        self.cross_attention_cache.reset()
        self.is_updated.clear()

    def batch_select_indices(self, indices: Tensor) -> None:
        self.self_attention_cache.batch_select_indices(indices)
        self.cross_attention_cache.batch_select_indices(indices)

    def batch_repeat_interleave(self, repeats: int) -> None:
        self.self_attention_cache.batch_repeat_interleave(repeats)
        self.cross_attention_cache.batch_repeat_interleave(repeats)

    def crop(self, max_length: int) -> None:
        self.self_attention_cache.crop(max_length)
        self.cross_attention_cache.crop(max_length)

    @override
    def get_max_cache_shape(self) -> int | None:
        return self.self_attention_cache.get_max_cache_shape()



from .alex import *
from .coatnet import *
from .convnext import *
from .crossvit import *
from .cspnet import *
from .cvt import *
from .dense import *
from .detr import *
from .efficient import *
from .efficientdet import *
from .efficientformer import *
from .fast_rcnn import *
from .faster_rcnn import *
from .inception_next import *
from .inception_res import *
from .inception import *
from .lenet import *
from .maxvit import *
from .mobile import *
from .pvt import *
from .rcnn import *
from .resnest import *
from .resnet import *
from .resnext import *
from .senet import *
from .sknet import *
from .swin import *
from .vgg import *
from .vit import *
from .xception import *
from .yolo import *
from .zfnet import *



import unicodedata

from collections import Counter
from pathlib import Path
from typing import Any, Iterable

from lucid.data.tokenizers import Tokenizer, SpecialTokens

from lucid._backend._C.tokenizers.core import _C_WordPieceTokenizer


__all__ = ["WordPieceTokenizer", "WordPieceTokenizerFast"]


class WordPieceTokenizer(Tokenizer):
    def __init__(
        self,
        vocab: dict[str, int] | None = None,
        vocab_file: Path | str | None = None,
        unk_token: SpecialTokens | str = SpecialTokens.UNK,
        pad_token: SpecialTokens | str = SpecialTokens.PAD,
        bos_token: SpecialTokens | str | None = None,
        eos_token: SpecialTokens | str | None = None,
        lowercase: bool = True,
        wordpieces_prefix: str = "##",
        max_input_chars_per_word: int = 100,
        clean_text: bool = True,
        verbose: bool = False,
    ) -> None:
        if vocab is not None and vocab_file is not None:
            raise ValueError("Provide only one of 'vocab' or 'vocab_file'.")

        super().__init__(
            unk_token=unk_token,
            pad_token=pad_token,
            bos_token=bos_token,
            eos_token=eos_token,
        )

        if vocab_file is not None:
            vocab = self._load_vocab(vocab_file)
        if vocab is None:
            vocab = {}

        self.lowercase = lowercase
        self.wordpieces_prefix = wordpieces_prefix
        self.max_input_chars_per_word = max_input_chars_per_word
        self.clean_text = clean_text

        self.vocab: dict[str, int] = dict(vocab)
        self._ensure_special_tokens()
        self.ids_to_tokens = self._build_id_to_token_map(self.vocab, verbose=verbose)

    def fit(
        self,
        texts: Iterable[str],
        vocab_size: int,
        min_frequency: int = 2,
        verbose: bool = False,
    ) -> WordPieceTokenizer:
        trained_vocab = self._train_vocab(
            texts=texts,
            vocab_size=vocab_size,
            min_frequency=min_frequency,
        )
        self.vocab = trained_vocab
        unk_token = self.unk_token or SpecialTokens.UNK.value

        self.ids_to_tokens = self._build_id_to_token_map(
            self.vocab, unk_token=unk_token, verbose=verbose
        )
        return self

    @classmethod
    def train_from_iterator(
        cls,
        texts: Iterable[str],
        vocab_size: int,
        min_frequency: int = 2,
        verbose: bool = False,
        **kwargs,
    ) -> WordPieceTokenizer:
        tokenizer = cls(**kwargs)
        tokenizer.fit(
            texts, vocab_size=vocab_size, min_frequency=min_frequency, verbose=verbose
        )
        return tokenizer

    @property
    def vocab_size(self) -> int:
        return len(self.vocab)

    def tokenize(self, text: str) -> list[str]:
        if self.clean_text:
            text = self._clean_text(text)

        basic_tokens = self._basic_tokenize(text)
        tokens: list[str] = []
        for token in basic_tokens:
            tokens.extend(self._wordpiece_tokenize(token))
        return tokens

    def convert_tokens_to_ids(self, tokens: str | list[str]) -> int | list[int]:
        unk_id = self._unk_id()
        if isinstance(tokens, str):
            return self.vocab.get(tokens, unk_id)
        return [self.vocab.get(token, unk_id) for token in tokens]

    def convert_ids_to_tokens(self, ids: int | list[int]) -> str | list[str]:
        if isinstance(ids, int):
            return self._id_to_token(ids)
        return [self._id_to_token(i) for i in ids]

    def convert_tokens_to_string(self, tokens: list[str]) -> str:
        out = " ".join(tokens)
        out = out.replace(f" {self.wordpieces_prefix}", "")

        for punct in [".", ",", "!", "?", ";", ":", "%", ")", "]", "}"]:
            out = out.replace(f" {punct}", punct)
        for punct in ["(", "[", "{"]:
            out = out.replace(f"{punct} ", punct)

        return out.strip()

    def save_pretrained(self, save_directory: Path | str) -> list[str]:
        save_path = Path(save_directory)
        save_path.mkdir(parents=True, exist_ok=True)

        vocab_path = save_path / "vocab.txt"
        config_path = save_path / "tokenizer_config.json"

        with vocab_path.open("w", encoding="utf-8") as f:
            for token in self.ids_to_tokens:
                f.write(token + "\n")

        config = {
            "unk_token": self.unk_token,
            "pad_token": self.pad_token,
            "bos_token": self.bos_token,
            "eos_token": self.eos_token,
            "lowercase": self.lowercase,
            "wordpieces_prefix": self.wordpieces_prefix,
            "max_input_chars_per_word": self.max_input_chars_per_word,
            "clean_text": self.clean_text,
        }
        self._save_tokenizer_config(save_path, config)

        return [str(vocab_path), str(config_path)]

    @classmethod
    def from_pretrained(
        cls, pretrained_model_name_or_path: str | Path, **kwrags
    ) -> WordPieceTokenizer:
        path = Path(pretrained_model_name_or_path)
        if path.is_dir():
            vocab_file = path / "vocab.txt"
        else:
            vocab_file = path

        if not vocab_file.exists():
            raise FileNotFoundError(f"Cannot find vocabulary file: {vocab_file}")

        allowed = {
            "unk_token",
            "pad_token",
            "bos_token",
            "eos_token",
            "lowercase",
            "wordpieces_prefix",
            "max_input_chars_per_word",
            "clean_text",
        }
        init_kwargs = cls._load_tokenizer_config(path, allowed_keys=allowed)

        init_kwargs.update(kwrags)
        return cls(vocab_file=vocab_file, **init_kwargs)

    @staticmethod
    def _load_vocab(vocab_file: Path | str) -> dict[str, int]:
        vocab: dict[str, int] = {}
        with Path(vocab_file).open("r", encoding="utf-8") as f:
            for idx, line in enumerate(f):
                token = line.rstrip("\n")
                if token:
                    vocab[token] = idx
        return vocab

    def _ensure_special_tokens(self) -> None:
        for token in self.all_special_tokens:
            if token not in self.vocab:
                self.vocab[token] = len(self.vocab)

    def _train_vocab(
        self,
        texts: Iterable[str],
        vocab_size: int,
        min_frequency: int,
    ) -> dict[str, int]:
        special_tokens = self.all_special_tokens
        if vocab_size < len(special_tokens):
            raise ValueError(
                f"'vocab_size' must be >= number of special tokens ({len(special_tokens)})."
            )
        if min_frequency < 1:
            raise ValueError("'min_frequency' must be >= 1.")

        word_freq = self._build_word_frequency(texts)
        if not word_freq:
            raise ValueError("Cannot train WordPiece vocabulary from empty corpus.")

        word_splits, token_set = self._initialize_splits(word_freq)
        merged_tokens: list[str] = []

        target_size = vocab_size - len(special_tokens)
        while len(token_set) + len(merged_tokens) < target_size:
            best_pair = self._select_best_pair(word_splits, word_freq)
            if best_pair is None:
                break
            merged = self._merge_pair(best_pair, word_splits)
            if merged not in token_set and merged not in merged_tokens:
                merged_tokens.append(merged)

        vocab_tokens = list(dict.fromkeys(sorted(token_set) + merged_tokens))
        vocab_tokens = vocab_tokens[:target_size]

        full_vocab = special_tokens + vocab_tokens
        return {token: i for i, token in enumerate(full_vocab)}

    def _build_word_frequency(self, texts: Iterable[str]) -> Counter[str]:
        word_freq: Counter[str] = Counter()
        for text in texts:
            if self.clean_text:
                text = self._clean_text(text)
            for token in self._basic_tokenize(text):
                if token:
                    word_freq[token] += 1
        return word_freq

    def _initialize_splits(
        self, word_freq: Counter[str]
    ) -> tuple[dict[str, list[str]], set[str]]:
        word_splits: dict[str, list[str]] = {}
        token_set: set[str] = set()

        for word in word_freq:
            if not word:
                continue
            split = [word[0]] + [self.wordpieces_prefix + ch for ch in word[1:]]
            word_splits[word] = split
            token_set.update(split)

        return word_splits, token_set

    def _select_best_pair(
        self, word_splits: dict[str, list[str]], word_freq: Counter[str]
    ) -> tuple[str, str] | None:
        token_freq: Counter[str] = Counter()
        pair_freq: Counter[tuple[str, str]] = Counter()

        for word, split in word_splits.items():
            freq = word_freq[word]
            for token in split:
                token_freq[token] += freq
            for i in range(len(split) - 1):
                pair_freq[(split[i], split[i + 1])] += freq

        best_pair: tuple[str, str] | None = None
        best_score = -1.0
        best_pair_count = -1

        for pair, pfreq in pair_freq.items():
            if pfreq < 1:
                continue
            left, right = pair
            denom = token_freq[left] * token_freq[right]
            if denom == 0:
                continue
            score = pfreq / denom
            if (
                score > best_score
                or (score == best_score and pfreq > best_pair_count)
                or (
                    score == best_score
                    and pfreq == best_pair_count
                    and pair < (best_pair or pair)
                )
            ):
                best_score = score
                best_pair_count = pfreq
                best_pair = pair

        if best_pair is None or best_pair_count < 1:
            return None
        return best_pair

    def _merge_pair(
        self, pair: tuple[str, str], word_splits: dict[str, list[str]]
    ) -> str:
        left, right = pair
        if right.startswith(self.wordpieces_prefix):
            merged_token = left + right[len(self.wordpieces_prefix) :]
        else:
            merged_token = left + right

        for word, split in word_splits.items():
            if len(split) < 2:
                continue
            new_split: list[str] = []
            i = 0
            while i < len(split):
                if i < len(split) - 1 and split[i] == left and split[i + 1] == right:
                    new_split.append(merged_token)
                    i += 2
                else:
                    new_split.append(split[i])
                    i += 1
            word_splits[word] = new_split

        return merged_token

    def _unk_id(self) -> int:
        unk = self.unk_token
        if unk is None:
            raise ValueError("'unk_token' must not be None for WordPiece tokenizer.")
        if unk not in self.vocab:
            raise ValueError(f"Unknown token '{unk}' is not in vocabulary.")
        return self.vocab[unk]

    def _id_to_token(self, idx: int) -> str:
        if idx < 0 or idx >= len(self.ids_to_tokens):
            return self.unk_token or SpecialTokens.UNK.value
        token = self.ids_to_tokens[idx]
        return token if token else (self.unk_token or SpecialTokens.UNK.value)

    def _basic_tokenize(self, text: str) -> list[str]:
        tokens: list[str] = []
        for token in text.strip().split():
            if self.lowercase:
                token = token.lower()
            tokens.extend(self._split_on_punctuation(token))
        return tokens

    @staticmethod
    def _split_on_punctuation(token: str) -> list[str]:
        out: list[str] = []
        current: list[str] = []
        for ch in token:
            if WordPieceTokenizer._is_punctuation(ch):
                if current:
                    out.append("".join(current))
                    current = []
                out.append(ch)
            else:
                current.append(ch)
        if current:
            out.append("".join(current))
        return out

    def _wordpiece_tokenize(self, token: str) -> list[str]:
        if len(token) > self.max_input_chars_per_word:
            return [self.unk_token or SpecialTokens.UNK.value]

        chars = list(token)
        start = 0
        sub_tokens: list[str] = []

        while start < len(chars):
            end = len(chars)
            cur_substr: str | None = None
            while start < end:
                substr = "".join(chars[start:end])
                if start > 0:
                    substr = self.wordpieces_prefix + substr
                if substr in self.vocab:
                    cur_substr = substr
                    break
                end -= 1

            if cur_substr is None:
                return [self.unk_token or SpecialTokens.UNK.value]

            sub_tokens.append(cur_substr)
            start = end

        return sub_tokens

    @staticmethod
    def _is_whitespace(ch: str) -> bool:
        if ch in (" ", "\t", "\n", "\r"):
            return True
        return unicodedata.category(ch) == "Zs"

    @staticmethod
    def _is_control(ch: str) -> bool:
        if ch in ("\t", "\n", "\r"):
            return False
        return unicodedata.category(ch) in {"Cc", "Cf"}

    @staticmethod
    def _is_punctuation(ch: str) -> bool:
        cp = ord(ch)
        if (
            (33 <= cp <= 47)
            or (58 <= cp <= 64)
            or (91 <= cp <= 96)
            or (123 <= cp <= 126)
        ):
            return True
        return unicodedata.category(ch).startswith("P")

    @staticmethod
    def _clean_text(text: str) -> str:
        out: list[str] = []
        for ch in text:
            cp = ord(ch)
            if cp in (0, 0xFFFD) or WordPieceTokenizer._is_control(ch):
                continue
            if WordPieceTokenizer._is_whitespace(ch):
                out.append(" ")
            else:
                out.append(ch)
        return "".join(out)


class WordPieceTokenizerFast(Tokenizer):
    def __init__(
        self,
        vocab: dict[str, int] | None = None,
        vocab_file: Path | str | None = None,
        unk_token: SpecialTokens | str = SpecialTokens.UNK,
        pad_token: SpecialTokens | str = SpecialTokens.PAD,
        bos_token: SpecialTokens | str | None = None,
        eos_token: SpecialTokens | str | None = None,
        lowercase: bool = True,
        wordpieces_prefix: str = "##",
        max_input_chars_per_word: int = 100,
        clean_text: bool = True,
    ) -> None:
        super().__init__(
            unk_token=unk_token,
            pad_token=pad_token,
            bos_token=bos_token,
            eos_token=eos_token,
        )

        self.lowercase = lowercase
        self.wordpieces_prefix = wordpieces_prefix
        self.max_input_chars_per_word = max_input_chars_per_word
        self.clean_text = clean_text

        if _C_WordPieceTokenizer is None:
            raise ImportError(
                "Cannot import native C++ WordPiece backend "
                "'lucid._backend._C.tokenizers.core'. "
                "Build extensions first."
            )

        self._backend = _C_WordPieceTokenizer(
            vocab=vocab,
            vocab_file=Path(vocab_file) if vocab_file is not None else None,
            unk_token=self.unk_token,
            pad_token=self.pad_token,
            bos_token=self.bos_token,
            eos_token=self.eos_token,
            wordpieces_prefix=self.wordpieces_prefix,
            max_input_chars_per_word=self.max_input_chars_per_word,
            lowercase=self.lowercase,
            clean_text=self.clean_text,
        )

    @property
    def vocab_size(self) -> int:
        size = getattr(self._backend, "vocab_size")
        return int(size() if callable(size) else size)

    def tokenize(self, text: str) -> list[str]:
        return list(self._backend.tokenize(text))

    def convert_tokens_to_ids(self, tokens: str | list[str]) -> int | list[int]:
        if isinstance(tokens, str):
            return int(self._backend.convert_token_to_id(tokens))
        return [int(x) for x in self._backend.convert_tokens_to_ids(tokens)]

    def convert_ids_to_tokens(self, ids: int | list[int]) -> str | list[str]:
        if isinstance(ids, int):
            return str(self._backend.convert_id_to_token(ids))
        return [str(x) for x in self._backend.convert_ids_to_tokens(ids)]

    def convert_tokens_to_string(self, tokens: list[str]) -> str:
        return str(self._backend.convert_tokens_to_string(tokens))

    def fit(
        self,
        texts: Iterable[str],
        vocab_size: int,
        min_frequency: int = 2,
        **_: Any,
    ) -> WordPieceTokenizerFast:
        self._backend.fit(list(texts), int(vocab_size), int(min_frequency))
        return self

    @classmethod
    def train_from_iterator(
        cls,
        texts: Iterable[str],
        vocab_size: int,
        min_frequency: int = 2,
        **kwargs: Any,
    ) -> WordPieceTokenizerFast:
        tokenizer = cls(**kwargs)
        tokenizer.fit(texts, vocab_size=vocab_size, min_frequency=min_frequency)
        return tokenizer

    def save_pretrained(self, save_directory: Path | str) -> list[str]:
        save_path = Path(save_directory)
        save_path.mkdir(parents=True, exist_ok=True)

        vocab_path = save_path / "vocab.txt"
        with vocab_path.open("w", encoding="utf-8") as f:
            for idx in range(self.vocab_size):
                f.write(self.convert_ids_to_tokens(idx) + "\n")

        config = {
            "unk_token": self.unk_token,
            "pad_token": self.pad_token,
            "bos_token": self.bos_token,
            "eos_token": self.eos_token,
            "lowercase": self.lowercase,
            "wordpieces_prefix": self.wordpieces_prefix,
            "max_input_chars_per_word": self.max_input_chars_per_word,
            "clean_text": self.clean_text,
        }
        config_path = self._save_tokenizer_config(save_path, config)
        return [str(vocab_path), str(config_path)]

    @classmethod
    def from_pretrained(
        cls, pretrained_model_name_or_path: str | Path, **kwrags: Any
    ) -> WordPieceTokenizerFast:
        path = Path(pretrained_model_name_or_path)
        vocab_file = path / "vocab.txt" if path.is_dir() else path
        if not vocab_file.exists():
            raise FileNotFoundError(f"Cannot find vocabulary file: {vocab_file}")

        allowed = {
            "unk_token",
            "pad_token",
            "bos_token",
            "eos_token",
            "lowercase",
            "wordpieces_prefix",
            "max_input_chars_per_word",
            "clean_text",
        }
        init_kwargs = cls._load_tokenizer_config(path, allowed_keys=allowed)
        init_kwargs.update(kwrags)
        return cls(vocab_file=vocab_file, **init_kwargs)



from dataclasses import dataclass
from typing import Iterable, Sequence

import lucid

from lucid._tensor import Tensor
from lucid.types import _Scalar


__all__ = [
    "PackedSequence",
    "pad_sequence",
    "pack_padded_sequence",
    "pad_packed_sequence",
    "pack_sequence",
    "unpack_sequence",
]


@dataclass(frozen=True)
class PackedSequence:
    data: Tensor
    batch_sizes: Tensor
    sorted_indices: Tensor | None = None
    unsorted_indices: Tensor | None = None


def pad_sequence(
    sequences: Iterable[Tensor], batch_first: bool = False, padding_value: _Scalar = 0
) -> Tensor:
    seq_list = list(sequences)
    if not seq_list:
        raise ValueError("pad_sequence expected a non-empty iterable of Tensors")

    first = seq_list[0]
    if not isinstance(first, Tensor):
        raise TypeError("pad_sequence expects Tensor elements")

    ndim = first.ndim
    if ndim < 1:
        raise ValueError("pad_sequence expects tensors with at least 1 dimension")

    trailing_shape = first.shape[1:]
    device = first.device
    dtype = first.dtype

    lengths: list[int] = []
    for idx, seq in enumerate(seq_list):
        if not isinstance(seq, Tensor):
            raise TypeError("pad_sequence expects Tensor elements")
        if seq.ndim != ndim:
            raise ValueError(
                f"pad_sequence expects tensors with {ndim} dimensions, "
                f"got {seq.ndim} at index {idx}"
            )
        if seq.shape[1:] != trailing_shape:
            raise ValueError(
                "pad_sequence expects all tensors to share the same trailing shape"
            )
        if seq.device != device:
            raise ValueError("pad_sequence expects all tensors on the same device")
        if seq.dtype != dtype:
            raise ValueError("pad_sequence expects all tensors with the same dtype")
        lengths.append(seq.shape[0])

    max_len = max(lengths)
    batch_size = len(seq_list)

    if batch_first:
        out_shape = (batch_size, max_len, *trailing_shape)
    else:
        out_shape = (max_len, batch_size, *trailing_shape)

    output = lucid.full(out_shape, padding_value, dtype=dtype, device=device)
    for i, seq in enumerate(seq_list):
        length = lengths[i]
        if length == 0:
            continue
        if batch_first:
            output[i, :length] = seq
        else:
            output[:length, i] = seq

    return output


def _as_lengths(lengths: Sequence[int] | Tensor, *, device: str) -> Tensor:
    if isinstance(lengths, Tensor):
        return lengths
    return Tensor(list(lengths), device=device)


def _invert_permutation(indices: Tensor) -> Tensor:
    return lucid.argsort(indices, axis=0)


def pack_padded_sequence(
    input_: Tensor,
    lengths: Sequence[int] | Tensor,
    batch_first: bool = False,
    enforce_sorted: bool = True,
) -> PackedSequence:
    if input_.ndim < 2:
        raise ValueError(
            f"pack_padded_sequence expected input with at least 2 dims, got {input_.ndim}"
        )

    if batch_first:
        input_ = input_.swapaxes(0, 1)

    seq_len, batch_size = input_.shape[0], input_.shape[1]
    lengths_t = _as_lengths(lengths, device=input_.device)
    if lengths_t.ndim != 1:
        raise ValueError("lengths must be a 1D sequence or tensor")
    if lengths_t.shape[0] != batch_size:
        raise ValueError(
            f"lengths size {lengths_t.shape[0]} does not match batch size {batch_size}"
        )

    sorted_indices = None
    unsorted_indices = None

    if enforce_sorted:
        sorted_lengths = lengths_t
    else:
        sorted_indices = lucid.argsort(lengths_t, descending=True, axis=0)
        unsorted_indices = _invert_permutation(sorted_indices)

        lengths_t = lengths_t[sorted_indices]
        input_ = input_[:, sorted_indices]
        sorted_lengths = lengths_t

    max_len = int(sorted_lengths[0].item())
    if max_len > seq_len:
        raise ValueError(
            f"lengths has max {max_len} but input has sequence length {seq_len}"
        )

    batch_sizes: list[int] = []
    chunks: list[Tensor] = []
    for t in range(max_len):
        bs = int((sorted_lengths > t).sum().item())
        batch_sizes.append(bs)
        if bs == 0:
            break
        chunks.append(input_[t, :bs])

    if not chunks:
        data = input_[:0]
    else:
        data = lucid.concatenate(tuple(chunks), axis=0)

    return PackedSequence(
        data=data,
        batch_sizes=Tensor(batch_sizes, device=input_.device),
        sorted_indices=sorted_indices,
        unsorted_indices=unsorted_indices,
    )


def pad_packed_sequence(
    sequence: PackedSequence, batch_first: bool = False, padding_value: _Scalar = 0
) -> tuple[Tensor, Tensor]:
    data = sequence.data
    batch_sizes = sequence.batch_sizes
    if batch_sizes.ndim != 1:
        raise ValueError("batch_sizes must be 1D")

    max_len = int(batch_sizes.shape[0])
    if max_len == 0:
        raise ValueError("batch_sizes must be non-empty")

    batch_size = int(batch_sizes[0].item())
    trailing_shape = data.shape[1:]

    if batch_first:
        out_shape = (batch_size, max_len, *trailing_shape)
    else:
        out_shape = (max_len, batch_size, *trailing_shape)

    output = lucid.full(out_shape, padding_value, dtype=data.dtype, device=data.device)

    lengths = [0] * batch_size
    offset = 0
    for t in range(max_len):
        bs = int(batch_sizes[t].item())
        if bs == 0:
            break

        chunk = data[offset : offset + bs]
        offset += bs
        for i in range(bs):
            lengths[i] += 1
        if batch_first:
            output[:bs, t] = chunk
        else:
            output[t, :bs] = chunk

    lengths_t = Tensor(lengths, device=data.device)
    if sequence.unsorted_indices is not None:
        if batch_first:
            output = output[sequence.unsorted_indices]
        else:
            output = output[:, sequence.unsorted_indices]
        lengths_t = lengths_t[sequence.unsorted_indices]

    return output, lengths_t


def pack_sequence(
    sequences: Iterable[Tensor], enforce_sorted: bool = True
) -> PackedSequence:
    seq_list = list(sequences)
    if not seq_list:
        raise ValueError("pack_sequence expected a non-empty iterable of Tensors")

    lengths = [seq.shape[0] for seq in seq_list]
    padded = pad_sequence(seq_list, batch_first=False, padding_value=0.0)
    return pack_padded_sequence(
        padded, lengths, batch_first=False, enforce_sorted=enforce_sorted
    )


def unpack_sequence(
    sequence: PackedSequence, batch_first: bool = False
) -> list[Tensor]:
    padded, lengths = pad_packed_sequence(
        sequence, batch_first=batch_first, padding_value=0.0
    )
    result: list[Tensor] = []
    for i, length in enumerate(lengths):
        l = int(length.item())
        if batch_first:
            result.append(padded[i, :l])
        else:
            result.append(padded[:l, i])
    return result



"""
Lucid BERT tokenizer C++ bindings
"""

from __future__ import annotations
import collections.abc
import typing

__all__: list[str] = list()

class _C_BERTEncodePairResult:
    def __init__(self) -> None: ...
    @property
    def attention_mask(self) -> list[int]: ...
    @attention_mask.setter
    def attention_mask(
        self, arg0: collections.abc.Sequence[typing.SupportsInt]
    ) -> None: ...
    @property
    def input_ids(self) -> list[int]: ...
    @input_ids.setter
    def input_ids(self, arg0: collections.abc.Sequence[typing.SupportsInt]) -> None: ...
    @property
    def token_type_ids(self) -> list[int]: ...
    @token_type_ids.setter
    def token_type_ids(
        self, arg0: collections.abc.Sequence[typing.SupportsInt]
    ) -> None: ...

class _C_BERTTokenizer:
    def __init__(
        self,
        vocab: collections.abc.Mapping[str, typing.SupportsInt] | None = None,
        vocab_file: os.PathLike | str | bytes | None = None,
        unk_token: str = "[UNK]",
        pad_token: str = "[PAD]",
        cls_token: str = "[CLS]",
        mask_token: str = "[MASK]",
        wordpieces_prefix: str = "##",
        max_input_chars_per_word: typing.SupportsInt = 100,
        lowercase: bool = True,
        clean_text: bool = True,
    ) -> None: ...
    def build_inputs_with_special_tokens(
        self, tokens: collections.abc.Sequence[str]
    ) -> list[str]: ...
    def build_inputs_with_special_tokens_pair(
        self,
        tokens_a: collections.abc.Sequence[str],
        tokens_b: collections.abc.Sequence[str],
    ) -> list[str]: ...
    def cls_token(self) -> str: ...
    def convert_id_to_token(self, id: typing.SupportsInt) -> str: ...
    def convert_ids_to_tokens(
        self, ids: collections.abc.Sequence[typing.SupportsInt]
    ) -> list[str]: ...
    def convert_token_to_id(self, token: str) -> int: ...
    def convert_tokens_to_ids(
        self, tokens: collections.abc.Sequence[str]
    ) -> list[int]: ...
    def convert_tokens_to_string(
        self, tokens: collections.abc.Sequence[str]
    ) -> str: ...
    def create_token_type_ids_from_sequences(
        self,
        tokens_a: collections.abc.Sequence[str],
        tokens_b: collections.abc.Sequence[str] | None = None,
    ) -> list[int]: ...
    def encode_plus(
        self, text_a: str, text_b: str | None = None
    ) -> _C_BERTEncodePairResult: ...
    def mask_token(self) -> str: ...
    def sep_token(self) -> str: ...
    def tokenize(self, text: str) -> list[str]: ...
    def vocab_size(self) -> int: ...



import math
from typing import Literal

from lucid._tensor import Tensor
from lucid.nn._kernel.pool import pool_nd_kernel


def avg_pool1d(
    input_: Tensor,
    kernel_size: int | tuple[int] = 1,
    stride: int | tuple[int] = 1,
    padding: int | tuple[int] = 0,
) -> Tensor:
    return pool_nd_kernel(kernel_size, stride, padding, mode="avg")(input_)


def avg_pool2d(
    input_: Tensor,
    kernel_size: int | tuple[int, int] = 1,
    stride: int | tuple[int, int] = 1,
    padding: int | tuple[int, int] = 0,
) -> Tensor:
    return pool_nd_kernel(kernel_size, stride, padding, mode="avg")(input_)


def avg_pool3d(
    input_: Tensor,
    kernel_size: int | tuple[int, int, int] = 1,
    stride: int | tuple[int, int, int] = 1,
    padding: int | tuple[int, int, int] = 0,
) -> Tensor:
    return pool_nd_kernel(kernel_size, stride, padding, mode="avg")(input_)


def max_pool1d(
    input_: Tensor,
    kernel_size: int | tuple[int] = 1,
    stride: int | tuple[int] = 1,
    padding: int | tuple[int] = 0,
) -> Tensor:
    return pool_nd_kernel(kernel_size, stride, padding, mode="max")(input_)


def max_pool2d(
    input_: Tensor,
    kernel_size: int | tuple[int, int] = 1,
    stride: int | tuple[int, int] = 1,
    padding: int | tuple[int, int] = 0,
) -> Tensor:
    return pool_nd_kernel(kernel_size, stride, padding, mode="max")(input_)


def max_pool3d(
    input_: Tensor,
    kernel_size: int | tuple[int, int, int] = 1,
    stride: int | tuple[int, int, int] = 1,
    padding: int | tuple[int, int, int] = 0,
) -> Tensor:
    return pool_nd_kernel(kernel_size, stride, padding, mode="max")(input_)


def adaptive_pool1d(
    input_: Tensor, output_size: int, avg_or_max: Literal["avg", "max"]
) -> Tensor:
    L = input_.shape[2]
    kernel_size = math.ceil(L / output_size)
    stride = math.floor(L / output_size)
    pad = max((output_size - 1) * stride + kernel_size - L, 0)

    padding = (pad // 2,)
    if avg_or_max == "avg":
        return avg_pool1d(input_, kernel_size, stride, padding)
    else:
        return max_pool1d(input_, kernel_size, stride, padding)


def adaptive_pool2d(
    input_: Tensor,
    output_size: tuple[int, int] | int,
    avg_or_max: Literal["avg", "max"],
) -> Tensor:
    if isinstance(output_size, int):
        output_size = (output_size, output_size)

    _, _, H, W = input_.shape
    target_H, target_W = output_size

    kernel_h = math.ceil(H / target_H)
    kernel_w = math.ceil(W / target_W)

    stride_h = math.floor(H / target_H)
    stride_w = math.floor(W / target_W)

    pad_h = max((target_H - 1) * stride_h + kernel_h - H, 0)
    pad_w = max((target_W - 1) * stride_w + kernel_w - W, 0)

    padding = (pad_h // 2, pad_w // 2)
    if avg_or_max == "avg":
        return avg_pool2d(input_, (kernel_h, kernel_w), (stride_h, stride_w), padding)
    else:
        return max_pool2d(input_, (kernel_h, kernel_w), (stride_h, stride_w), padding)


def adaptive_pool3d(
    input_: Tensor,
    output_size: tuple[int, int, int] | int,
    avg_or_max: Literal["avg", "max"],
) -> Tensor:
    if isinstance(output_size, int):
        output_size = (output_size, output_size, output_size)

    _, _, D, H, W = input_.shape
    target_D, target_H, target_W = output_size

    kernel_d = math.ceil(D / target_D)
    kernel_h = math.ceil(H / target_H)
    kernel_w = math.ceil(W / target_W)

    stride_d = math.floor(D / target_D)
    stride_h = math.floor(H / target_H)
    stride_w = math.floor(W / target_W)

    pad_d = max((target_D - 1) * stride_d + kernel_d - D, 0)
    pad_h = max((target_H - 1) * stride_h + kernel_h - H, 0)
    pad_w = max((target_W - 1) * stride_w + kernel_w - W, 0)

    padding = (pad_d // 2, pad_h // 2, pad_w // 2)
    if avg_or_max == "avg":
        return avg_pool2d(
            input_,
            (kernel_d, kernel_h, kernel_w),
            (stride_d, stride_h, stride_w),
            padding,
        )
    else:
        return max_pool2d(
            input_,
            (kernel_d, kernel_h, kernel_w),
            (stride_d, stride_h, stride_w),
            padding,
        )
